import json
from copy import copy

import pytest
import responses

from datarobot import CustomInferenceModel, TARGET_TYPE
from datarobot.utils import underscorize

from tests.model.utils import assert_custom_model_version


@pytest.fixture
def mocked_models(mocked_custom_model_version, mocked_custom_model_version2):
    return {
        "count": 2,
        "totalCount": 2,
        "next": None,
        "data": [
            {
                "deploymentsCount": 0,
                "updated": "2020-04-30T09:55:38.319786Z",
                "supportsBinaryClassification": False,
                "description": "",
                "language": "",
                "created": "2020-04-30T09:55:38.319558Z",
                "supportsRegression": True,
                "createdBy": "admin@datarobot.com",
                "latestVersion": mocked_custom_model_version,
                "customModelType": "learning",
                "id": "5eaaa09a40240308176cee55",
                "name": "dddd",
            },
            {
                "multiseriesIdColumns": None,
                "supportsRegression": True,
                "trainingDatasetVersionId": "5ea6bbc7402403321d4e1fae",
                "timeSteps": None,
                "latestVersion": mocked_custom_model_version2,
                "targetName": "MEDV",
                "id": "5ea6b5da402403181895cc51",
                "featureDerivationWindowEnd": None,
                "datetimePartitionColumn": None,
                "trainingDatasetId": "5ea6bbc7402403321d4e1fad",
                "trainingDataPartitionColumn": None,
                "positiveClassLabel": "1",
                "customModelType": "inference",
                "updated": "2020-04-27T10:37:14.341673Z",
                "supportsBinaryClassification": False,
                "description": "",
                "negativeClassLabel": "0",
                "createdBy": "admin@datarobot.com",
                "timeUnit": None,
                "predictionThreshold": 0.5,
                "trainingDataFileName": "boston_housing.csv",
                "forecastWindowStart": None,
                "deploymentsCount": 0,
                "name": "ee",
                "language": "",
                "created": "2020-04-27T10:37:14.340932Z",
                "forecastWindowEnd": None,
                "isTimeSeries": False,
                "trainingDataAssignmentInProgress": False,
                "featureDerivationWindowStart": None,
            },
        ],
        "previous": None,
    }


@pytest.fixture
def mocked_learning_model(mocked_models):
    return mocked_models["data"][0]


@pytest.fixture
def mocked_inference_model(mocked_models):
    return mocked_models["data"][1]


@pytest.fixture
def mocked_inference_model_no_latest_version(mocked_inference_model):
    model = copy(mocked_inference_model)
    model["latestVersion"] = None
    return model


def assert_custom_model_common(model, model_json):
    assert model.id == model_json["id"]
    assert model.name == model_json["name"]
    assert model.description == model_json["description"]
    assert (
        (model.target_type == TARGET_TYPE.BINARY) == model_json["supportsBinaryClassification"]
    )
    assert (
        (model.target_type == TARGET_TYPE.REGRESSION) == model_json["supportsRegression"]
    )
    if model_json["latestVersion"]:
        assert_custom_model_version(model.latest_version, model_json["latestVersion"])
    else:
        assert model.latest_version is None
    assert model.deployments_count == model_json["deploymentsCount"]
    assert model.created_by == model_json["createdBy"]
    assert model.updated_at == model_json["updated"]
    assert model.created_at == model_json["created"]


def assert_inference_model(model, model_json):
    assert_custom_model_common(model, model_json)
    assert model.language == model_json["language"]
    assert model.target_name == model_json["targetName"]
    assert model.training_dataset_id == model_json["trainingDatasetId"]
    assert model.training_dataset_version_id == model_json["trainingDatasetVersionId"]
    assert model.training_data_file_name == model_json["trainingDataFileName"]
    assert (
        model.training_data_assignment_in_progress == model_json["trainingDataAssignmentInProgress"]
    )
    assert model.positive_class_label == model_json["positiveClassLabel"]
    assert model.negative_class_label == model_json["negativeClassLabel"]
    assert model.prediction_threshold == model_json["predictionThreshold"]
    assert model.training_data_partition_column == model_json["trainingDataPartitionColumn"]


@pytest.fixture
def make_models_url(unittest_endpoint):
    def _make_models_url(model_id=None):
        base_url = "{}/customModels/".format(unittest_endpoint)
        if model_id is not None:
            return "{}{}/".format(base_url, model_id)
        return base_url

    return _make_models_url


def mock_get_response(url, response):
    responses.add(
        responses.GET,
        url,
        status=200,
        content_type="application/json",
        body=json.dumps(response),
    )


class TestCustomInferenceModel(object):

    @pytest.mark.parametrize(
        "model_data",
        ["mocked_inference_model", "mocked_inference_model_no_latest_version"]
    )
    def test_from_server_data(self, request, model_data):
        model_data = request.getfixturevalue(model_data)
        model = CustomInferenceModel.from_server_data(model_data)
        assert_inference_model(model, model_data)

    @responses.activate
    def test_get_version(self, mocked_inference_model, make_models_url):
        # arrange
        url = make_models_url(mocked_inference_model["id"])
        mock_get_response(url, mocked_inference_model)

        # act
        model = CustomInferenceModel.get(mocked_inference_model["id"])

        # assert
        assert_inference_model(model, mocked_inference_model)

        assert responses.calls[0].request.method == "GET"
        assert responses.calls[0].request.url == url

    @responses.activate
    def test_list_versions(self, make_models_url, mocked_inference_model):
        # arrange
        url = make_models_url() + "?customModelType=inference"
        mock_get_response(
            url,
            {
                "count": 1,
                "totalCount": 1,
                "next": None,
                "previous": None,
                "data": [mocked_inference_model],
            },
        )

        # act
        models = CustomInferenceModel.list()

        # assert
        assert len(models) == 1
        assert_inference_model(models[0], mocked_inference_model)

        assert responses.calls[0].request.method == "GET"
        assert responses.calls[0].request.url == url

    @responses.activate
    def test_update(self, make_models_url, mocked_inference_model):
        # arrange
        attrs = {"description": "xx", "predictionThreshold": 0.1}

        url = make_models_url(mocked_inference_model["id"])
        mock_get_response(url, mocked_inference_model)

        mocked_inference_model.update(attrs)

        responses.add(
            responses.PATCH,
            url,
            status=200,
            content_type="application/json",
            body=json.dumps(mocked_inference_model),
        )

        # act
        model = CustomInferenceModel.get(mocked_inference_model["id"])
        model.update(**{underscorize(k): v for k, v in attrs.items()})

        # assert
        assert_inference_model(model, mocked_inference_model)

        assert responses.calls[0].request.method == "GET"
        assert responses.calls[0].request.url == url
        assert responses.calls[1].request.method == "PATCH"
        assert responses.calls[1].request.url == url
        assert responses.calls[1].request.body == json.dumps(attrs).encode()

    @responses.activate
    def test_refresh(self, make_models_url, mocked_inference_model):
        # arrange
        url = make_models_url(mocked_inference_model["id"])
        mock_get_response(url, mocked_inference_model)

        mocked_inference_model.update({"description": "xx", "predictionThreshold": 0.1})

        mock_get_response(url, mocked_inference_model)

        # act
        model = CustomInferenceModel.get(mocked_inference_model["id"])
        model.refresh()

        # assert
        assert_inference_model(model, mocked_inference_model)

        assert responses.calls[0].request.method == "GET"
        assert responses.calls[0].request.url == url
        assert responses.calls[1].request.method == "GET"
        assert responses.calls[1].request.url == url

    @responses.activate
    def test_delete(self, make_models_url, mocked_inference_model):
        url = make_models_url(mocked_inference_model["id"])

        mock_get_response(url, mocked_inference_model)
        responses.add(
            responses.DELETE, url, status=204, content_type="application/json",
        )

        model = CustomInferenceModel.get(mocked_inference_model["id"])
        model.delete()

        assert responses.calls[0].request.method == "GET"
        assert responses.calls[0].request.url == url
        assert responses.calls[1].request.method == "DELETE"
        assert responses.calls[1].request.url == url

    @responses.activate
    def test_download(self, make_models_url, mocked_inference_model, tmpdir):
        # arrange
        url = make_models_url(mocked_inference_model["id"])
        mock_get_response(url, mocked_inference_model)
        responses.add(
            responses.GET,
            url + "download/",
            status=200,
            content_type="application/json",
            body=b"content",
        )

        downloaded_file = tmpdir.mkdir("sub").join("download")
        downloaded_file_path = str(downloaded_file)

        # act
        model = CustomInferenceModel.get(mocked_inference_model["id"])
        model.download_latest_version(downloaded_file_path)

        # assert
        downloaded_file.read() == b"content"

        assert responses.calls[0].request.method == "GET"
        assert responses.calls[0].request.url == url
        assert responses.calls[1].request.method == "GET"
        assert responses.calls[1].request.url == url + "download/"

    @responses.activate
    def test_assign_training_data_non_blocking(
        self, unittest_endpoint, make_models_url, mocked_inference_model, tmpdir
    ):
        status_url = "{}/status_url".format(unittest_endpoint)

        mocked_inference_model["trainingDataAssignmentInProgress"] = False
        responses.add(
            responses.GET,
            make_models_url(mocked_inference_model["id"]),
            status=200,
            content_type="application/json",
            body=json.dumps(mocked_inference_model),
        )
        responses.add(
            responses.PATCH,
            make_models_url(mocked_inference_model["id"]) + "trainingData/",
            status=200,
            content_type="application/json",
            headers={"Location": status_url},
            body=json.dumps({"id": mocked_inference_model["id"]}),
        )
        mocked_inference_model["trainingDataAssignmentInProgress"] = True
        responses.add(
            responses.GET,
            make_models_url(mocked_inference_model["id"]),
            status=200,
            content_type="application/json",
            body=json.dumps(mocked_inference_model),
        )

        model = CustomInferenceModel.get(mocked_inference_model["id"])

        assert not model.training_data_assignment_in_progress
        model.assign_training_data("5ea6bbc7402403321d4e1fae", max_wait=None)

        assert model.training_data_assignment_in_progress

        assert responses.calls[0].request.method == "GET"
        assert responses.calls[0].request.url == make_models_url(
            mocked_inference_model["id"]
        )
        assert responses.calls[1].request.method == "PATCH"
        assert responses.calls[1].request.url == make_models_url(
            mocked_inference_model["id"]
        ) + "trainingData/"
        assert responses.calls[2].request.method == "GET"
        assert responses.calls[2].request.url == make_models_url(
            mocked_inference_model["id"]
        )

    @responses.activate
    def test_assign_training_data_blocking(
        self, unittest_endpoint, make_models_url, mocked_inference_model, tmpdir
    ):
        status_url = "{}/status_url".format(unittest_endpoint)

        mocked_inference_model["trainingDataAssignmentInProgress"] = False
        responses.add(
            responses.GET,
            make_models_url(mocked_inference_model["id"]),
            status=200,
            content_type="application/json",
            body=json.dumps(mocked_inference_model),
        )
        responses.add(
            responses.PATCH,
            make_models_url(mocked_inference_model["id"]) + "trainingData/",
            status=200,
            content_type="application/json",
            headers={"Location": status_url},
            body=json.dumps({"id": mocked_inference_model["id"]}),
        )
        responses.add(
            responses.GET,
            status_url,
            status=303,
            content_type="application/json",
            headers={"Location": status_url},
        )
        responses.add(
            responses.GET,
            make_models_url(mocked_inference_model["id"]),
            status=200,
            content_type="application/json",
            body=json.dumps(mocked_inference_model),
        )

        model = CustomInferenceModel.get(mocked_inference_model["id"])

        assert not model.training_data_assignment_in_progress
        model.assign_training_data("5ea6bbc7402403321d4e1fae")

        assert not model.training_data_assignment_in_progress

        assert responses.calls[0].request.method == "GET"
        assert responses.calls[0].request.url == make_models_url(
            mocked_inference_model["id"]
        )
        assert responses.calls[1].request.method == "PATCH"
        assert responses.calls[1].request.url == make_models_url(
            mocked_inference_model["id"]
        ) + "trainingData/"
        assert responses.calls[2].request.method == "GET"
        assert responses.calls[2].request.url == status_url
        assert responses.calls[3].request.method == "GET"
        assert responses.calls[3].request.url == make_models_url(
            mocked_inference_model["id"]
        )

    @responses.activate
    @pytest.mark.parametrize(
        "model_data",
        ["mocked_inference_model", "mocked_inference_model_no_latest_version"]
    )
    def test_create(self, request, make_models_url, model_data):
        model_data = request.getfixturevalue(model_data)
        responses.add(
            responses.POST,
            make_models_url(),
            status=200,
            content_type="application/json",
            body=json.dumps(model_data),
        )

        model = CustomInferenceModel.create(
            model_data["name"],
            TARGET_TYPE.REGRESSION,
            model_data["targetName"],
        )
        assert_inference_model(model, model_data)

        assert responses.calls[0].request.method == "POST"
        assert responses.calls[0].request.url == make_models_url()

    @responses.activate
    def test_create_copy(self, make_models_url, mocked_inference_model):
        custom_model_to_copy = "5ea6bbc7402403321d4e1fad"

        url = make_models_url() + "fromCustomModel/"
        responses.add(
            responses.POST,
            url,
            status=200,
            content_type="application/json",
            body=json.dumps(mocked_inference_model),
        )

        model = CustomInferenceModel.copy_custom_model(custom_model_to_copy)
        assert_inference_model(model, mocked_inference_model)

        req = responses.calls[0].request
        assert req.method == "POST"
        assert req.url == url
        assert json.loads(req.body) == {"customModelId": custom_model_to_copy}



import pytest
import responses

from datarobot import Client
from datarobot.client import get_client


def test_missing_token_in_env_fails(clean_env):
    clean_env.setenv('DATAROBOT_ENDPOINT', 'https://home.com')
    with pytest.raises(ValueError):
        get_client()


def test_missing_endpoint_in_config_fails():
    with pytest.raises(ValueError):
        Client(token='a-token')


def test_missing_token_in_config_fails():
    with pytest.raises(ValueError):
        Client(endpoint='https://host_name.com')


@pytest.mark.usefixtures('no_client_version_check')
def test_file_config_with_connect_timeout(clean_user_config, default_config_token,
                                          default_config_endpoint):
    fcontent = 'endpoint: {}\ntoken: {}\nconnect_timeout: 23'.format(
        default_config_endpoint,
        default_config_token
    )
    clean_user_config.write(fcontent)
    client = get_client()
    assert client.connect_timeout == 23


@pytest.mark.usefixtures('no_client_version_check')
def test_code_config_with_connect_timeout(code_config_token, code_config_endpoint):
    Client(token=code_config_token, endpoint=code_config_endpoint,
           connect_timeout=24)
    c = get_client()
    assert c.connect_timeout == 24


@responses.activate
def test_bad_token_message():
    endpoint = 'https://host_name.com/api/v2'
    responses.add(
        responses.GET,
        endpoint + '/version/',
        status=401,
    )
    with pytest.raises(UserWarning) as userwarning:
        Client(token='not correct', endpoint=endpoint)
    assert str(userwarning.value) == ('Unable to authenticate to the server - are you sure the '
                                      'provided token of "not correct" and endpoint of "https://'
                                      'host_name.com/api/v2" are correct? Note that if you access '
                                      'the DataRobot webapp at `https://app.datarobot.com`, then '
                                      'the correct endpoint to specify would be '
                                      '`https://app.datarobot.com/api/v2`.')


@responses.activate
def test_bad_endpoint_message():
    endpoint = 'https://host_name.com/api/v16'
    responses.add(
        responses.GET,
        endpoint + '/version/',
        status=404,
    )
    with pytest.raises(UserWarning) as userwarning:
        Client(token='not correct', endpoint=endpoint)
    assert str(userwarning.value) == ('Error retrieving a version string from the server. Server '
                                      'did not reply with an API version. This may indicate the '
                                      'endpoint parameter `https://host_name.com/api/v16` is '
                                      'incorrect, or that the server there is incompatible with '
                                      'this version of the DataRobot client package. Note that if '
                                      'you access the DataRobot webapp at `https://app.datarobot'
                                      '.com`, then the correct endpoint to specify would be '
                                      '`https://app.datarobot.com/api/v2`.')



import json

import pytest
import responses
from datarobot.errors import ClientError

from datarobot import Model

from datarobot.models.confusion_chart import ConfusionChart


@pytest.fixture
def confusion_chart_validation_data(model_id, confusion_chart_data):
    return {
        'source': 'validation',
        'sourceModelId': model_id,
        'data': confusion_chart_data
    }


@pytest.fixture
def confusion_chart_parent_model_validation_data(parent_model_id, confusion_chart_data):
    return {
        'source': 'validation',
        'sourceModelId': parent_model_id,
        'data': confusion_chart_data
    }


@pytest.fixture
def confusion_chart_parent_model_holdout_data(parent_model_id, confusion_chart_data):
    return {
        'source': 'holdout',
        'sourceModelId': parent_model_id,
        'data': confusion_chart_data
    }


@pytest.fixture
def confusion_chart_validation_metadata_server_data():
    return {
        'source': 'validation',
        'classNames': ['1', '2', '3'],
        'totalMatrixSum': 24,
        'relevantClassesPositions': [[1]]
    }


@pytest.fixture
def confusion_chart_holdout_metadata_server_data():
    return {
        'source': 'holdout',
        'classNames': ['1', '2', '3'],
        'totalMatrixSum': 24,
        'relevantClassesPositions': [[1]]
    }


def test_instantiation(confusion_chart_validation_data):
    confusion_chart_validation_data['data']['classes'] = ['1', '2', '3']
    confusion_chart = ConfusionChart.from_server_data(confusion_chart_validation_data)

    assert confusion_chart.classes == ['1', '2', '3']
    assert confusion_chart.confusion_matrix == [[9, 0, 0], [0, 4, 1], [0, 1, 9]]
    assert confusion_chart.source == 'validation'
    assert confusion_chart.source_model_id == confusion_chart_validation_data['sourceModelId']


def test_future_proof(confusion_chart_validation_data):
    data_with_future_keys = dict(confusion_chart_validation_data, new_key='some future data')
    data_with_future_keys['data']['classMetrics'][0]['new_key'] = 'some future bin data'
    data_with_future_keys['data']['classes'] = ['1', '2', '3']
    confusion_chart = ConfusionChart.from_server_data(data_with_future_keys)

    assert confusion_chart.classes == ['1', '2', '3']
    assert confusion_chart.confusion_matrix == [[9, 0, 0], [0, 4, 1], [0, 1, 9]]
    assert confusion_chart.source == 'validation'


@pytest.fixture
def confusion_chart_validation_data_url(project_id, model_id):
    return 'https://host_name.com/projects/{}/models/{}/confusionCharts/{}/'.format(
        project_id,
        model_id,
        'validation'
    )


@pytest.fixture
def confusion_chart_parent_model_validation_data_url(project_id, parent_model_id):
    return 'https://host_name.com/projects/{}/models/{}/confusionCharts/{}/'.format(
        project_id,
        parent_model_id,
        'validation'
    )


@pytest.fixture
def confusion_chart_list_url(project_id, model_id):
    return 'https://host_name.com/projects/{}/models/{}/confusionCharts/'.format(
        project_id,
        model_id
    )


@pytest.fixture
def confusion_chart_parent_list_url(project_id, parent_model_id):
    return 'https://host_name.com/projects/{}/models/{}/confusionCharts/'.format(
        project_id,
        parent_model_id
    )


@pytest.fixture
def frozen_model_url(project_id, model_id):
    return 'https://host_name.com/projects/{}/frozenModels/{}/'.format(
        project_id,
        model_id
    )


@pytest.fixture
def confusion_chart_validation_metadata_response(project_url, model_id,
                                                 confusion_chart_validation_metadata_server_data):
    responses.add(
        responses.GET,
        '{}models/{}/confusionCharts/validation/metadata/'.format(project_url, model_id),
        json=confusion_chart_validation_metadata_server_data, content_type='application/json'
    )


@pytest.fixture
def confusion_chart_parent_model_validation_metadata_response(
        project_url, parent_model_id, confusion_chart_validation_metadata_server_data):
    responses.add(
        responses.GET,
        '{}models/{}/confusionCharts/validation/metadata/'.format(project_url, parent_model_id),
        json=confusion_chart_validation_metadata_server_data, content_type='application/json'
    )


@pytest.fixture
def confusion_chart_parent_model_holdout_metadata_response(
        project_url, parent_model_id, confusion_chart_validation_metadata_server_data):
    responses.add(
        responses.GET,
        '{}models/{}/confusionCharts/holdout/metadata/'.format(project_url, parent_model_id),
        json=confusion_chart_validation_metadata_server_data, content_type='application/json'
    )


@responses.activate
@pytest.mark.usefixtures('confusion_chart_validation_metadata_response')
def test_get_validation_confusion_chart(confusion_chart_validation_data,
                                        confusion_chart_validation_data_url,
                                        project_id, model_id):
    responses.add(
        responses.GET,
        confusion_chart_validation_data_url,
        status=200,
        content_type='application/json',
        body=json.dumps(confusion_chart_validation_data)
    )
    model = Model(id=model_id, project_id=project_id)
    cm = model.get_confusion_chart('validation')

    assert cm.source == confusion_chart_validation_data['source']
    assert cm.classes == ['1', '2', '3']
    assert cm.confusion_matrix == [[9, 0, 0], [0, 4, 1], [0, 1, 9]]
    assert cm.source_model_id == model_id


@responses.activate
def test_get_frozen_validation_confusion_chart_no_fallback(confusion_chart_validation_data_url,
                                                           project_id, model_id):
    responses.add(
        responses.GET,
        confusion_chart_validation_data_url,
        status=404
    )
    model = Model(id=model_id, project_id=project_id, is_frozen=True)
    with pytest.raises(ClientError):
        model.get_confusion_chart('validation')


@responses.activate
@pytest.mark.usefixtures('confusion_chart_parent_model_validation_metadata_response',
                         'known_warning')
def test_get_frozen_validation_confusion_chart_with_fallback(
        confusion_chart_parent_model_validation_data,
        confusion_chart_validation_data_url,
        confusion_chart_parent_model_validation_data_url,
        frozen_model_url,
        project_id, model_id, parent_model_id,
        frozen_json):

    responses.add(
        responses.GET,
        confusion_chart_validation_data_url,
        status=404
    )
    responses.add(
        responses.GET,
        frozen_model_url,
        body=frozen_json,
        status=200,
        content_type='application/json'
    )
    responses.add(
        responses.GET,
        confusion_chart_parent_model_validation_data_url,
        status=200,
        content_type='application/json',
        body=json.dumps(confusion_chart_parent_model_validation_data)
    )
    model = Model(id=model_id, project_id=project_id, is_frozen=True)
    cm = model.get_confusion_chart('validation', fallback_to_parent_insights=True)

    assert cm.source == confusion_chart_parent_model_validation_data['source']
    assert cm.classes == ['1', '2', '3']
    assert cm.confusion_matrix == [[9, 0, 0], [0, 4, 1], [0, 1, 9]]
    assert cm.source_model_id == parent_model_id


@responses.activate
@pytest.mark.usefixtures('confusion_chart_validation_metadata_response')
def test_get_all_confusion_chart(confusion_chart_validation_data,
                                 confusion_chart_list_url,
                                 project_id, model_id):
    responses.add(
        responses.GET,
        confusion_chart_list_url,
        status=200,
        content_type='application/json',
        body=json.dumps({'charts': [confusion_chart_validation_data]})
    )
    model = Model(id=model_id, project_id=project_id)
    cm_list = model.get_all_confusion_charts()

    assert len(cm_list) == 1
    assert cm_list[0].source == confusion_chart_validation_data['source']
    assert cm_list[0].classes == ['1', '2', '3']
    assert cm_list[0].confusion_matrix == [[9, 0, 0], [0, 4, 1], [0, 1, 9]]
    assert cm_list[0].source_model_id == model_id


@responses.activate
@pytest.mark.usefixtures('confusion_chart_validation_metadata_response')
def test_get_frozen_all_confusion_chart_no_fallback(confusion_chart_validation_data,
                                                    confusion_chart_list_url,
                                                    project_id, model_id):
    responses.add(
        responses.GET,
        confusion_chart_list_url,
        status=200,
        content_type='application/json',
        body=json.dumps({'charts': [confusion_chart_validation_data]})
    )
    model = Model(id=model_id, project_id=project_id, is_frozen=True)
    cm_list = model.get_all_confusion_charts()

    assert len(cm_list) == 1
    assert cm_list[0].source == confusion_chart_validation_data['source']
    assert cm_list[0].classes == ['1', '2', '3']
    assert cm_list[0].confusion_matrix == [[9, 0, 0], [0, 4, 1], [0, 1, 9]]
    assert cm_list[0].source_model_id == model_id


@responses.activate
@pytest.mark.usefixtures('confusion_chart_validation_metadata_response',
                         'confusion_chart_parent_model_validation_metadata_response',
                         'confusion_chart_parent_model_holdout_metadata_response',
                         'known_warning')
def test_get_frozen_all_confusion_chart_with_fallback(confusion_chart_validation_data,
                                                      confusion_chart_parent_model_validation_data,
                                                      confusion_chart_parent_model_holdout_data,
                                                      confusion_chart_list_url,
                                                      confusion_chart_parent_list_url,
                                                      frozen_model_url,
                                                      project_id, model_id, parent_model_id,
                                                      frozen_json):
    responses.add(
        responses.GET,
        confusion_chart_list_url,
        status=200,
        content_type='application/json',
        body=json.dumps({'charts': [confusion_chart_validation_data]})
    )
    responses.add(
        responses.GET,
        frozen_model_url,
        body=frozen_json,
        status=200,
        content_type='application/json'
    )
    responses.add(
        responses.GET,
        confusion_chart_parent_list_url,
        status=200,
        content_type='application/json',
        body=json.dumps({'charts': [confusion_chart_parent_model_validation_data,
                                    confusion_chart_parent_model_holdout_data]})
    )
    model = Model(id=model_id, project_id=project_id, is_frozen=True)
    cm_list = model.get_all_confusion_charts(fallback_to_parent_insights=True)

    assert len(cm_list) == 2
    assert cm_list[0].source == confusion_chart_validation_data['source']
    assert cm_list[0].classes == ['1', '2', '3']
    assert cm_list[0].confusion_matrix == [[9, 0, 0], [0, 4, 1], [0, 1, 9]]
    assert cm_list[0].source_model_id == model_id

    assert cm_list[1].source == confusion_chart_parent_model_holdout_data['source']
    assert cm_list[1].classes == ['1', '2', '3']
    assert cm_list[1].confusion_matrix == [[9, 0, 0], [0, 4, 1], [0, 1, 9]]
    assert cm_list[1].source_model_id == parent_model_id



import six
import trafaret as t

from datarobot import errors
from datarobot.models.api_object import APIObject


class SecondaryDataset(object):
    """A secondary dataset to be used for feature discovery

    .. versionadded:: v2.20

    Attributes
    ----------
    identifier: str
        a short name of this secondary dataset (will be used as part of the generated feature names)
    catalog_id: str
        catalog Id of the secondary dataset
    catalog_version_id: str
        catalog version Id of the secondary dataset
    snapshot_policy: str (default: `latest`)
        policy to be used by the secondary dataset when making predictions
        Must be one of the following values:
            'specified': Use specific snapshot specified by catalogVersionId
            'latest': Use latest snapshot from the same catalog item
            'dynamic': Get data from the source (only applicable for JDBC datasets)
    """

    def __init__(
        self, identifier, catalog_id, catalog_version_id, snapshot_policy='latest'
    ):
        self.identifier = identifier
        self.catalog_id = catalog_id
        self.catalog_version_id = catalog_version_id
        self.snapshot_policy = snapshot_policy

    def to_dict(self):
        return {
            'identifier': self.identifier,
            'catalogId': self.catalog_id,
            'catalogVersionId': self.catalog_version_id,
            'snapshotPolicy': self.snapshot_policy,
        }


class DatasetConfiguration(object):
    """Specify a dataset configuration

    .. versionadded:: v2.20

    Attributes
    ----------
    feature_engineering_graph_id: str
        id of the feature engineering graph
    secondary_datasets: list of SecondaryDataset
        list of secondary datasets

    """

    def __init__(self, feature_engineering_graph_id=None, secondary_datasets=None):
        self.feature_engineering_graph_id = feature_engineering_graph_id
        self.secondary_datasets = secondary_datasets

    def to_dict(self):
        return {
            'featureEngineeringGraphId': self.feature_engineering_graph_id,
            'secondaryDatasets': [dataset.to_dict() for dataset in self.secondary_datasets],
        }


class SecondaryDatasetConfigurations(APIObject):
    """ Create secondary dataset configurations for a given project

    .. versionadded:: v2.20

    Attributes
    ----------
    id : str
        id of this secondary dataset configuration
    project_id : str
        id of the associated project.
    config: list of DatasetConfiguration
        list of secondary dataset configurations
    """

    _base_url = 'projects/{}/secondaryDatasetsConfigurations/'
    _secondary_dataset_converter = t.Dict(
        {
            t.Key('identifier'): t.String(min_length=3, max_length=20),
            t.Key('catalog_version_id'): t.String,
            t.Key('catalog_id'): t.String,
            t.Key('snapshot_policy', optional=True, default='latest'): t.Enum(
                'latest', 'specified', 'dynamic')
        }
    )

    _dataset_configuration_converter = t.Dict(
        {
            t.Key('feature_engineering_graph_id'): t.String(),
            t.Key('secondary_datasets'): t.List(_secondary_dataset_converter),
        }
    ).ignore_extra('*')

    _converter = t.Dict(
        {
            t.Key('id'): t.String(),
            t.Key('project_id'): t.String(),
            t.Key('config'): t.List(_dataset_configuration_converter),
        }
    ).ignore_extra('*')

    def __init__(self, id=None, project_id=None, config=None):
        self.id = id
        self.project_id = project_id
        self.config = config

    def to_dict(self):
        return {
            'id': self.id,
            'project_id': self.project_id,
            'config': [c.to_dict() for c in self.config],
        }

    @classmethod
    def from_data(cls, data):
        checked = cls._converter.check(data)
        safe_data = cls._filter_data(checked)

        id = safe_data.get('id', None)
        project_id = safe_data.get('project_id', None)
        dataset_configs = safe_data.get('config', None)
        conf_list = []
        for conf in dataset_configs:
            graph_id = conf.get('feature_engineering_graph_id', None)
            datasets = conf.get('secondary_datasets', None)
            dataset_list = []
            for d in datasets:
                dataset = SecondaryDataset(**d)
                dataset_list.append(dataset)
            dataset_configuration = DatasetConfiguration(graph_id, dataset_list)
            conf_list.append(dataset_configuration)
        return SecondaryDatasetConfigurations(id, project_id, conf_list)

    @classmethod
    def create(cls, project_id, dataset_configurations):
        """ create secondary dataset configurations

        .. versionadded:: v2.20

        Parameters
        ----------
        project_id : str
            id of the associated project.
        dataset_configurations: list of DatasetConfiguration
            list of dataset configurations

        Returns
        -------
        an instance of SecondaryDatasetConfigurations

        Raises
        ------
        ClientError
            raised if incorrect configuration parameters are provided
        """
        if not project_id:
            raise errors.ClientError(six.text_type('project_id cannot be None or empty'), 422)
        if not dataset_configurations:
            raise errors.ClientError(
                six.text_type('dataset_configurations cannot be None or empty'),
                422,
            )

        url = cls._base_url.format(project_id)
        payload = {'conf': [conf.to_dict() for conf in dataset_configurations]}
        response = cls._client.post(url, data=payload)

        status = response.status_code
        if status == 201:
            return cls.from_server_data(response.json())
        else:
            error_msg = response.json().get(
                'message', six.text_type('error in processing secondary dataset configuration')
            )
            raise errors.ClientError(
                error_msg + six.text_type(' with server returned status {}').format(status),
                status,
            )

    def delete(self):
        """ Removes the Secondary datasets configuration

        .. versionadded:: v2.21

        Raises
        ------
        ClientError
            Raised if an invalid or already deleted secondary dataset config id is provided

        Examples
        --------
        .. code-block:: python

            # Deleting with a valid secondary_dataset_config id
            status_code = dr.SecondaryDatasetConfigurations.delete(some_config_id)
            status_code
            >>> 204
        """
        url = self._base_url.format(self.project_id)
        self._client.delete('{}{}/'.format(url, self.id))

    def get(self):
        """ Retrieve a single secondary dataset configuration for a given id

        .. versionadded:: v2.21

        Returns
        -------
        secondary_dataset_configurations : SecondaryDatasetConfigurations
            The requested secondary dataset configurations
        """
        url = self._base_url.format(self.project_id)
        return self.from_location('{}{}/'.format(url, self.id))



import json

import pytest
import responses

from datarobot import Project


@pytest.fixture
def project(project_with_target_data):
    return Project.from_data(project_with_target_data)


@pytest.fixture
def feature_association_details_data():
    return {
        "values": [
            [
                168,
                33,
                0.045
            ],
            [
                1,
                1,
                0.329
            ],
            [
                37,
                17,
                0.005
            ],
            [
                207,
                41,
                0.014
            ],
            [
                178,
                35,
                0.006
            ],
            [
                34,
                16,
                0.032
            ],
            [
                194,
                37,
                0.004
            ]
        ],
        "features": [
            "area1",
            "area2"
        ],
        "types": [
            "N",
            "N"
        ]
    }


@pytest.fixture
def feature_association_featurelists():
    return {
        "featurelists": [
            {
                "featurelistId": "54e510ef8bd88f5aeb02a3ed",
                "hasFam": True,
                "title": "Informative Features"
            }
        ]
    }


@pytest.fixture
def feature_association_list_url(project_id):
    base_url = 'https://host_name.com/projects/{}/featureAssociationMatrixDetails/' \
        .format(project_id)
    base_url += '?feature1=area1&feature2=area2'
    return base_url


@pytest.fixture
def feature_association_featurelist_url(project_id):
    url = 'https://host_name.com/projects/{}/featureAssociationFeaturelists/' \
        .format(project_id)
    return url


@responses.activate
def test_get_feature_association_details(feature_association_details_data,
                                         feature_association_list_url,
                                         project, project_id):
    responses.add(
        responses.GET,
        feature_association_list_url,
        status=200,
        content_type='application/json',
        body=json.dumps(feature_association_details_data)
    )
    feature_values = project.get_association_matrix_details(feature1='area1', feature2='area2')

    assert len(feature_values.keys()) == 3
    assert len(feature_values['values']) == 7
    assert feature_values['features'] == ['area1', 'area2']


@responses.activate
def test_get_feature_association_featurelists(feature_association_featurelists,
                                              feature_association_featurelist_url,
                                              project, project_id):
    responses.add(
        responses.GET,
        feature_association_featurelist_url,
        status=200,
        content_type='application/json',
        body=json.dumps(feature_association_featurelists)
    )
    featurelists = project.get_association_featurelists()

    assert len(featurelists.keys()) == 1
    assert len(featurelists['featurelists']) == 1
    assert featurelists['featurelists'][0]['title'] == 'Informative Features'



import json

import pytest
import responses


@pytest.fixture()
def deployments_data():
    deployments = [{
        'id': '5c6f7a9e9ca9b20017ff95a2', 'label': 'is_bad Predictions', 'description': '',
        'model': {'projectId': '5c0a969859b00004ba52e41b', 'targetName': 'is_bad',
                  'id': '5cb9c8b979872d014a0bb080', 'type': 'TensorFlow Deep Learning Classifier'},
        'defaultPredictionServer': {'id': '5a7c7a72192765001fc22769',
                                    'url': 'https://pred.example.com', 'datarobot-key': '27782'},
        'predictionUsage': {'dailyRates': [5, 2, 10, 5, 13, 4, 7],
                            'lastTimestamp': '2019-04-22T04:03:45.000000Z'},
        'capabilities': {'supportsModelReplacement': True, 'supportsDriftTracking': True},
        'permissions': [
            'CAN_SHARE', 'CAN_VIEW', 'CAN_DELETE_DEPLOYMENT', 'CAN_MAKE_PREDICTIONS'
        ],
        'serviceHealth': {'status': 'unknown'},
        'modelHealth': {'status': 'unknown',
                        'message': '1 important feature is drifting',
                        'startDate': '2019-04-18T02:00:00.000000Z',
                        'endDate': '2019-04-22T02:00:00.000000Z'},
        'accuracyHealth': {'status': 'unavailable', 'startDate': None, 'endDate': None}
    }, {
        'id': '5c6f7a9e9ca9b20017ff93b7', 'label': 'is_bad Predictions2', 'description': None,
        'model': {'projectId': '5c0a969859b00004ba52e41b', 'targetName': 'is_bad',
                  'id': '5cb9c8b979872d014a0bb080', 'type': 'TensorFlow Deep Learning Classifier'},
        'defaultPredictionServer': {'id': '5a7c7a72192765001fc22769',
                                    'url': 'https://pred.example.com', 'datarobot-key': '27782'},
        'predictionUsage': {'dailyRates': [], 'lastTimestamp': None},
        'capabilities': {'supportsModelReplacement': False, 'supportsDriftTracking': False},
        'permissions': [
            'CAN_SHARE', 'CAN_VIEW', 'CAN_DELETE_DEPLOYMENT', 'CAN_MAKE_PREDICTIONS'
        ],
        'serviceHealth': {'status': 'unknown'},
        'modelHealth': {'status': 'unknown', 'startDate': None, 'endDate': None},
        'accuracyHealth': {'status': 'unavailable', 'startDate': None, 'endDate': None}
    }]
    return deployments


@pytest.fixture()
def deployment_data(deployments_data):
    return deployments_data[0]


@pytest.fixture()
def deployment_get_response(unittest_endpoint, deployments_data):
    for deployment_data in deployments_data:
        url = '{}/{}/{}/'.format(unittest_endpoint, 'deployments', deployment_data['id'])
        responses.add(
            responses.GET,
            url,
            status=200,
            content_type='application/json',
            body=json.dumps(deployment_data)
        )



import datetime
import threading
import time
import warnings

import trafaret as t

from datarobot import errors
from datarobot.models.job import AbstractSpecificJob
from datarobot.models.credential import Credential
from datarobot.utils import (
    get_id_from_response,
    recognize_sourcedata,
    pagination,
    to_api
)
from ..enums import JOB_TYPE, QUEUE_STATUS, AVAILABLE_STATEMENT_TYPES
from ..utils import parse_time


class BatchPredictionJob(AbstractSpecificJob):
    """
    A Batch Prediction Job is used to score large data sets on
    prediction servers using the Batch Prediction API.

    Attributes
    ----------
    id : str
        the id of the job
    """
    _job_spec = t.Dict({
        t.Key('num_concurrent'): t.Int(),
        t.Key('threshold_high', optional=True): t.Float(),
        t.Key('threshold_low', optional=True): t.Float(),
        t.Key('deployment_id'): t.String(),
        t.Key('passthrough_columns', optional=True): t.List(t.String(allow_blank=True)),
        t.Key('passthrough_columns_set', optional=True): t.String(),
        t.Key('max_explanations', optional=True): t.Int(),
        t.Key('prediction_warning_enabled', optional=True): t.Bool(),
        t.Key('intake_settings', optional=True): t.Dict().allow_extra("*"),
        t.Key('output_settings', optional=True): t.Dict().allow_extra("*"),
        t.Key('timeseries_settings', optional=True): t.Dict().allow_extra("*"),
    }).allow_extra('*')
    _links = t.Dict({
        t.Key('download', optional=True): t.String(allow_blank=True),
        t.Key('self'): t.String(),
    }).allow_extra('*')
    _converter_extra = t.Dict({
        t.Key('percentage_completed'): t.Float(),
        t.Key('elapsed_time_sec'): t.Int(),
        t.Key('links'): _links,
        t.Key('job_spec'): _job_spec,
        t.Key('status_details'): t.String()
    }).allow_extra('*')

    _converter_common = t.Dict({
        t.Key('id', optional=True): t.String,
        t.Key('status', optional=True): t.Enum(
            QUEUE_STATUS.ABORTED,
            QUEUE_STATUS.COMPLETED,
            QUEUE_STATUS.ERROR,
            QUEUE_STATUS.INPROGRESS,
            QUEUE_STATUS.QUEUE,
            QUEUE_STATUS.RUNNING,
            QUEUE_STATUS.INITIALIZING),
        t.Key('project_id', optional=True): t.String,
        t.Key('is_blocked', optional=True): t.Bool,
    })

    _s3_settings = t.Dict({
        t.Key('url'): t.String(),
        t.Key('credential_id', optional=True): t.String(),
    })

    _jdbc_intake_settings = t.Dict({
        t.Key('data_store_id'): t.String(),
        t.Key('query', optional=True): t.String(),
        t.Key('table', optional=True): t.String(),
        t.Key('schema', optional=True): t.String(),
        t.Key('fetch_size', optional=True): t.Int(),
        t.Key('credential_id', optional=True): t.String(),
    })

    _jdbc_output_settings = t.Dict({
        t.Key('data_store_id'): t.String(),
        t.Key('table'): t.String(),
        t.Key('schema', optional=True): t.String(),
        t.Key('statement_type'): t.Enum(
            AVAILABLE_STATEMENT_TYPES.INSERT,
            AVAILABLE_STATEMENT_TYPES.UPDATE,
            AVAILABLE_STATEMENT_TYPES.INSERT_UPDATE,
            AVAILABLE_STATEMENT_TYPES.CREATE_TABLE
        ),
        t.Key('update_columns', optional=True): t.List(t.String),
        t.Key('where_columns', optional=True): t.List(t.String),
        t.Key('credential_id', optional=True): t.String(),
    })

    _timeseries_settings = t.Dict({
        t.Key('type'): t.Atom('forecast'),
        t.Key('forecast_point', optional=True): parse_time,
        t.Key('relax_known_in_advance_features_check', optional=True): t.Bool(),
    }) | t.Dict({
        t.Key('type'): t.Atom('historical'),
        t.Key('predictions_start_date', optional=True): parse_time,
        t.Key('predictions_end_date', optional=True): parse_time,
        t.Key('relax_known_in_advance_features_check', optional=True): t.Bool(),
    })

    _csv_settings = t.Dict({
        t.Key('delimiter', optional=True): t.Atom('tab') | t.String(min_length=1, max_length=1),
        t.Key('quotechar', optional=True): t.String(),
        t.Key('encoding', optional=True): t.String(),
    })

    _prediction_instance = t.Dict({
        t.Key('hostName'): t.String(),
        t.Key('sslEnabled', optional=True): t.Bool(),
        t.Key('datarobotKey', optional=True): t.String(),
        t.Key('apiKey', optional=True): t.String(),
    })

    _column_names_remapping = t.Mapping(t.String, t.String | t.Null())

    @classmethod
    def _job_type(cls):
        return JOB_TYPE.BATCH_PREDICTIONS

    @classmethod
    def _jobs_path(cls):
        return 'batchPredictions/'

    @classmethod
    def _job_path(cls, project_id, batch_prediction_job_id):
        return 'batchPredictions/{}/'.format(batch_prediction_job_id)

    @classmethod
    def _from_existing_path(cls):
        return 'batchPredictions/fromExisting/'

    @classmethod
    def score(
        cls,
        deployment,
        intake_settings=None,
        output_settings=None,
        csv_settings=None,
        timeseries_settings=None,
        num_concurrent=None,
        passthrough_columns=None,
        passthrough_columns_set=None,
        max_explanations=None,
        threshold_high=None,
        threshold_low=None,
        prediction_warning_enabled=None,
        include_prediction_status=False,
        skip_drift_tracking=False,
        prediction_instance=None,
        abort_on_error=True,
        column_names_remapping=None,
        include_probabilities=True,
        include_probabilities_classes=None,
    ):
        """
        Create new batch prediction job, upload the scoring dataset and
        return a batch prediction job.

        The default intake and output options are both `localFile` which
        requires the caller to pass the `file` parameter and either
        download the results using the `download()` method afterwards or
        pass a path to a file where the scored data will be downloaded to
        afterwards.

        Attributes
        ----------
        deployment : Deployment or string ID
            Deployment which will be used for scoring.

        intake_settings : dict (optional)
            A dict configuring how data is coming from. Supported options:

                - type : string, either `localFile`, `s3`, `azure`, `gcp`, `dataset` or `jdbc`

            To score from a local file, add the this parameter to the
            settings:

                - file : file-like object, string path to file or a
                  pandas.DataFrame of scoring data

            To score from S3, add the next parameters to the settings:

                - url : string, the URL to score (e.g.: `s3://bucket/key`)
                - credential_id : string (optional)

            .. _batch_predictions_jdbc_creds_usage:

            To score from JDBC, add the next parameters to the settings:

                - data_store_id : string, the ID of the external data store connected
                  to the JDBC data source (see
                  :ref:`Database Connectivity <database_connectivity_overview>`).
                - query : string (optional if `table` and `schema` is specified),
                  a self-supplied SELECT statement of the data set you wish to predict.
                - table : string (optional if `query` is specified),
                  the name of specified database table.
                - schema : string (optional if `query` is specified),
                  the name of specified database schema.
                - fetch_size : int (optional),
                  Changing the `fetchSize` can be used to balance throughput and memory
                  usage.
                - credential_id : string (optional) the ID of the credentials holding
                  information about a user with read-access to the JDBC data source (see
                  :ref:`Credentials <credentials_api_doc>`).

        output_settings : dict (optional)
            A dict configuring how scored data is to be saved. Supported
            options:

                - type : string, either `localFile`, `s3` or `jdbc`

            To save scored data to a local file, add this parameters to the
            settings:

                - path : string (optional), path to save the scored data
                  as CSV. If a path is not specified, you must download
                  the scored data yourself with `job.download()`.
                  If a path is specified, the call will block until the
                  job is done. if there are no other jobs currently
                  processing for the targeted prediction instance,
                  uploading, scoring, downloading will happen in parallel
                  without waiting for a full job to complete. Otherwise,
                  it will still block, but start downloading the scored
                  data as soon as it starts generating data. This is the
                  fastest method to get predictions.

            To save scored data to S3, add the next parameters to the settings:

                - url : string, the URL for storing the results
                  (e.g.: `s3://bucket/key`)
                - credential_id : string (optional)

            To save scored data to JDBC, add the next parameters to the settings:

                - `data_store_id` : string, the ID of the external data store connected to
                  the JDBC data source (see
                  :ref:`Database Connectivity <database_connectivity_overview>`).
                - `table` : string,  the name of specified database table.
                - `schema` : string (optional), the name of specified database schema.
                - `statement_type` : string, the type of insertion statement to create,
                  one of ``datarobot.enums.AVAILABLE_STATEMENT_TYPES``.
                - `update_columns` : list(string) (optional),  a list of strings containing
                  those column names to be updated in case `statement_type` is set to a
                  value related to update or upsert.
                - `where_columns` : list(string) (optional), a list of strings containing
                  those column names to be selected in case `statement_type` is set to a
                  value related to insert or update.
                - `credential_id` : string, the ID of the credentials holding information about
                  a user with write-access to the JDBC data source (see
                  :ref:`Credentials <credentials_api_doc>`).

        csv_settings : dict (optional)
            CSV intake and output settings. Supported options:

            - `delimiter` : string (optional, default `,`), fields are delimited by
              this character. Use the string `tab` to denote TSV (TAB separated values).
              Must be either a one-character string or the string `tab`.
            - `quotechar` : string (optional, default `"`), fields containing the
              delimiter must be quoted using this character.
            - `encoding` : string (optional, default `utf-8`), encoding for the CSV
              files. For example (but not limited to): `shift_jis`, `latin_1` or
              `mskanji`.

        timeseries_settings : dict (optional)
            Configuration for time-series scoring. Supported options:

            - `type` : string, must be `forecast` or `historical` (default if
              not passed is `forecast`). `forecast` mode makes predictions using
              `forecast_point` or rows in the dataset without target. `historical`
              enables bulk prediction mode which calculates predictions for all
              possible forecast points and forecast distances in the dataset within
              `predictions_start_date`/`predictions_end_date` range.
            - `forecast_point` : datetime (optional), forecast point for the dataset,
              used for the forecast predictions, by default value will be inferred
              from the dataset. May be passed if ``timeseries_settings.type=forecast``.
            - `predictions_start_date` : datetime (optional), used for historical
              predictions in order to override date from which predictions should be
              calculated. By default value will be inferred automatically from the
              dataset. May be passed if ``timeseries_settings.type=historical``.
            - `predictions_end_date` : datetime (optional), used for historical
              predictions in order to override date from which predictions should be
              calculated. By default value will be inferred automatically from the
              dataset. May be passed if ``timeseries_settings.type=historical``.
            - `relax_known_in_advance_features_check` : bool, (default `False`).
              If True, missing values in the known in advance features are allowed
              in the forecast window at the prediction time. If omitted or False,
              missing values are not allowed.

            .. warning:: This is an early release **beta feature**. While the API is stable,
              we are still working on ensuring the best performance possible of
              the scoring pipeline.

        num_concurrent : int (optional)
            Number of concurrent chunks to score simultaneously. Defaults to
            the available number of cores of the deployment. Lower it to leave
            resources for real-time scoring.

        passthrough_columns : list[string] (optional)
            Keep these columns from the scoring dataset in the scored dataset.
            This is useful for correlating predictions with source data.

        passthrough_columns_set : string (optional)
            To pass through every column from the scoring dataset, set this to
            `all`. Takes precedence over `passthrough_columns` if set.

        max_explanations : int (optional)
            Compute prediction explanations for this amount of features.

        threshold_high : float (optional)
            Only compute prediction explanations for predictions above this
            threshold. Can be combined with `threshold_low`.

        threshold_low : float (optional)
            Only compute prediction explanations for predictions below this
            threshold. Can be combined with `threshold_high`.

        prediction_warning_enabled : boolean (optional)
            Add prediction warnings to the scored data. Currently only
            supported for regression models.

        include_prediction_status : boolean (optional)
            Include the prediction_status column in the output, defaults to `False`.

        skip_drift_tracking : boolean (optional)
            Skips drift tracking on any predictions made from this job. This is useful when running
            non-production workloads to not affect drift tracking and cause unnecessary alerts.
            Defaults to `False`.

        prediction_instance : dict (optional)
            Defaults to instance specified by deployment or system configuration.
            Supported options:

                - `hostName` : string
                - `sslEnabled` : boolean (optional, default `true`). Set to `false` to
                  run prediction requests from the batch prediction job without SSL.
                - `datarobotKey` : string (optional), if running a job against a prediction
                  instance in the Managed AI Cloud, you must provide the organization level
                  DataRobot-Key
                - `apiKey` : string (optional), by default, prediction requests will use the
                  API key of the user that created the job. This allows you to make requests
                  on behalf of other users.

        abort_on_error : boolean (optional)
             Default behaviour is to abort the job if too many rows fail scoring. This will free
             up resources for other jobs that may score successfully. Set to `false` to
             unconditionally score every row no matter how many errors are encountered.
             Defaults to `True`.

        column_names_remapping : dict (optional)
            Mapping with column renaming for output table. Defaults to `{}`.

        include_probabilities : boolean (optional)
            Flag that enables returning of all probability columns. Defaults to `True`.

        include_probabilities_classes : list (optional)
            List the subset of classes if a user doesn't want all the classes. Defaults to `[]`.

        Returns
        -------
        BatchPredictionJob
            Instance of BatchPredictonJob
        """
        try:
            dep_id = deployment.id
        except AttributeError:
            dep_id = deployment

        job_data = {'deploymentId': dep_id}

        if num_concurrent is not None:
            job_data['numConcurrent'] = int(num_concurrent)

        if max_explanations is not None:
            job_data['maxExplanations'] = int(max_explanations)

        if threshold_high is not None:
            job_data['thresholdHigh'] = float(threshold_high)

        if threshold_low is not None:
            job_data['thresholdLow'] = float(threshold_low)

        if include_prediction_status:
            job_data['includePredictionStatus'] = bool(include_prediction_status)

        if skip_drift_tracking:
            job_data['skipDriftTracking'] = bool(skip_drift_tracking)

        if not abort_on_error:
            job_data['abortOnError'] = bool(abort_on_error)

        if not include_probabilities:
            job_data['includeProbabilities'] = bool(include_probabilities)

        if prediction_warning_enabled is not None:
            job_data['predictionWarningEnabled'] = \
                bool(prediction_warning_enabled)

        if passthrough_columns_set == 'all':
            job_data['passthroughColumnsSet'] = 'all'
        elif passthrough_columns is not None:
            job_data['passthroughColumns'] = passthrough_columns

        if prediction_instance is not None:
            cls._prediction_instance.check(prediction_instance)
            job_data['predictionInstance'] = to_api(prediction_instance)

        if column_names_remapping is not None:
            cls._column_names_remapping.check(column_names_remapping)
            job_data['columnNamesRemapping'] = dict(column_names_remapping)

        if include_probabilities_classes:
            job_data['includeProbabilitiesClasses'] = list(include_probabilities_classes)

        if timeseries_settings is not None:
            warnings.warn(
                'Time Series Batch Predictions an early release beta feature. '
                'While the API is stable, we are still working on ensuring the '
                'best performance possible of the scoring pipeline.'
            )
            timeseries_settings = cls._timeseries_settings.check(timeseries_settings)

            if timeseries_settings['type'] == 'historical':

                start_date = timeseries_settings.get('predictions_start_date')
                end_date = timeseries_settings.get('predictions_end_date')

                if start_date and not isinstance(start_date, datetime.datetime):
                    raise ValueError(
                        'The value provided for predictions_start_date was '
                        'not a valid format.'
                    )

                if end_date and not isinstance(end_date, datetime.datetime):
                    raise ValueError(
                        'The value provided for predictions_end_date was '
                        'not a valid format.'
                    )

                if start_date and not end_date:
                    raise ValueError(
                        'You must also provide predictions_end_date if you '
                        'provide predictions_start_date'
                    )

                if start_date and start_date > end_date:
                    raise ValueError(
                        'The value provided for predictions_end_date must be after '
                        'the value provided for predictions_start_date'
                    )

            if timeseries_settings['type'] == 'forecast':

                forecast_point = timeseries_settings.get('forecast_point')

                if forecast_point and not isinstance(forecast_point, datetime.datetime):
                    raise ValueError(
                        'The value provided for forecast_point was '
                        'not a valid format.'
                    )

            job_data['timeseriesSettings'] = to_api(timeseries_settings)

        if intake_settings is None:
            intake_settings = {'type': 'localFile'}

        intake_file = None

        # Validate the intake settings

        if intake_settings.get('type') not in ('localFile', 'azure', 'gcp', 's3', 'jdbc'):

            raise ValueError('Unsupported type parameter for intake_settings: %s' % (
                intake_settings.get('type'),
            ))

        elif intake_settings['type'] == 'localFile':

            # This intake option requires us to upload the source
            # data ourselves

            if intake_settings.get('file') is None:

                raise ValueError(
                    'Missing source data. Either supply the `file` '
                    'parameter or switch to an intake option that does not '
                    'require it.'
                )

            intake_file = recognize_sourcedata(
                intake_settings.pop('file'),
                'prediction.csv',
            )

        elif intake_settings['type'] == 's3':

            del intake_settings['type']
            intake_settings = cls._s3_settings.check(intake_settings)
            intake_settings['type'] = 's3'

        elif intake_settings['type'] == 'jdbc':

            del intake_settings['type']
            intake_settings = cls._jdbc_intake_settings.check(intake_settings)
            intake_settings['type'] = 'jdbc'

        job_data['intakeSettings'] = to_api(intake_settings)

        if output_settings is None:
            output_settings = {'type': 'localFile'}

        output_file = None

        # Validate the output settings

        if output_settings.get('type') not in ('localFile', 'azure', 'gcp', 's3', 'jdbc'):

            raise ValueError('Unsupported type parameter for output_settings: %s' % (
                output_settings.get('type'),
            ))

        elif output_settings['type'] == 'localFile':

            if output_settings.get('path') is not None:
                output_file = open(output_settings.pop('path'), 'wb')

        elif output_settings['type'] == 's3':

            del output_settings['type']
            output_settings = cls._s3_settings.check(output_settings)
            output_settings['type'] = 's3'

        elif output_settings['type'] == 'jdbc':

            del output_settings['type']
            output_settings = cls._jdbc_output_settings.check(output_settings)
            output_settings['type'] = 'jdbc'

        job_data['outputSettings'] = to_api(output_settings)

        if csv_settings is not None:
            cls._csv_settings.check(csv_settings)
            job_data['csvSettings'] = to_api(csv_settings)

        response = cls._client.post(
            url=cls._jobs_path(),
            json=job_data,
        )

        job_response = response.json()
        job_id = get_id_from_response(response)

        upload_thread = None

        if intake_file is not None:

            # There is source data to upload, so spin up a thread to handle
            # the upload concurrently and for thread safety issues, make
            # a copy of the REST client object

            _upload_client = cls._client.copy()

            def _uploader():

                if 'file_path' in intake_file:
                    fileobj = open(intake_file['file_path'], 'rb')
                else:
                    fileobj = intake_file['filelike']
                    fileobj.seek(0)

                try:
                    _upload_client.put(
                        url=job_response['links']['csvUpload'],
                        data=fileobj,
                        headers={'content-type': 'text/csv'},
                    )
                finally:
                    if hasattr(fileobj, 'close'):
                        fileobj.close()

            if output_file is not None:

                # If output_file is specified, we upload and download
                # concurrently

                upload_thread = threading.Thread(target=_uploader)
                upload_thread.setDaemon(True)
                upload_thread.start()

            else:

                # Otherwise, upload is synchronous

                _uploader()

        job = BatchPredictionJob.get(job_id)

        if output_file is not None:

            # We must download the result to `output_file`
            # And clean up any thread we spawned during uploading
            try:
                job.download(output_file)
            finally:
                output_file.close()
                if upload_thread is not None:
                    upload_thread.join()

        return job

    @classmethod
    def score_to_file(cls, deployment, intake_path, output_path, **kwargs):
        """
        Create new batch prediction job, upload the scoring dataset and
        download the scored CSV file concurrently.

        Will block until the entire file is scored.

        Refer to the `create` method for details on the other `kwargs`
        parameters.

        Attributes
        ----------
        deployment : Deployment or string ID
            Deployment which will be used for scoring.

        intake_path : file-like object/string path to file/pandas.DataFrame
              Scoring data

        output_path : str
            Filename to save the result under

        Returns
        -------
        BatchPredictionJob
            Instance of BatchPredictonJob
        """
        kwargs['intake_settings'] = {
            'type': 'localFile',
            'file': intake_path,
        }

        kwargs['output_settings'] = {
            'type': 'localFile',
            'path': output_path,
        }

        return cls.score(deployment, **kwargs)

    @classmethod
    def score_s3(
        cls,
        deployment,
        source_url,
        destination_url,
        credential=None,
        **kwargs
    ):
        """
        Create new batch prediction job, with a scoring dataset from S3
        and writing the result back to S3.

        This returns immediately after the job has been created. You
        must poll for job completion using `get_status()` or
        `wait_for_completion()`.

        Refer to the `create` method for details on the other `kwargs`
        parameters.

        Attributes
        ----------
        deployment : Deployment or string ID
            Deployment which will be used for scoring.

        source_url : string
            The URL for the prediction dataset (e.g.: `s3://bucket/key`)

        destination_url : string
            The URL for the scored dataset (e.g.: `s3://bucket/key`)

        credential : string or Credential (optional)
            The AWS Credential object or credential id

        Returns
        -------
        BatchPredictionJob
            Instance of BatchPredictonJob
        """
        intake_settings = {
            'type': 's3',
        }

        if credential is not None:
            if isinstance(credential, Credential):
                intake_settings['credential_id'] = credential.credential_id
            else:
                intake_settings['credential_id'] = credential

        output_settings = intake_settings.copy()

        intake_settings['url'] = source_url
        output_settings['url'] = destination_url

        kwargs.update({
            'intake_settings': intake_settings,
            'output_settings': output_settings,
        })

        return cls.score(deployment, **kwargs)

    @classmethod
    def score_azure(
        cls,
        deployment,
        source_url,
        destination_url,
        credential=None,
        **kwargs
    ):
        """
        Create new batch prediction job, with a scoring dataset from Azure blob
        storage and writing the result back to Azure blob storage.

        This returns immediately after the job has been created. You
        must poll for job completion using `get_status()` or
        `wait_for_completion()`.

        Refer to the `create` method for details on the other `kwargs`
        parameters.

        Attributes
        ----------
        deployment : Deployment or string ID
            Deployment which will be used for scoring.

        source_url : string
            The URL for the prediction dataset
            (e.g.: `https://storage_account.blob.endpoint/container/blob_name`)

        destination_url : string
            The URL for the scored dataset
            (e.g.: `https://storage_account.blob.endpoint/container/blob_name`)

        credential : string or Credential (optional)
            The Azure Credential object or credential id

        Returns
        -------
        BatchPredictionJob
            Instance of BatchPredictonJob
        """
        intake_settings = {
            'type': 'azure',
        }

        if credential is not None:
            if isinstance(credential, Credential):
                intake_settings['credential_id'] = credential.credential_id
            else:
                intake_settings['credential_id'] = credential

        output_settings = intake_settings.copy()

        intake_settings['url'] = source_url
        output_settings['url'] = destination_url

        kwargs.update({
            'intake_settings': intake_settings,
            'output_settings': output_settings,
        })

        return cls.score(deployment, **kwargs)

    @classmethod
    def score_gcp(
        cls,
        deployment,
        source_url,
        destination_url,
        credential=None,
        **kwargs
    ):
        """
        Create new batch prediction job, with a scoring dataset from Google Cloud Storage
        and writing the result back to one.

        This returns immediately after the job has been created. You
        must poll for job completion using `get_status()` or
        `wait_for_completion()`.

        Refer to the `create` method for details on the other `kwargs`
        parameters.

        Attributes
        ----------
        deployment : Deployment or string ID
            Deployment which will be used for scoring.

        source_url : string
            The URL for the prediction dataset
            (e.g.: `http(s)://storage.googleapis.com/[bucket]/[object]`)

        destination_url : string
            The URL for the scored dataset
            (e.g.: `http(s)://storage.googleapis.com/[bucket]/[object]`)

        credential : string or Credential (optional)
            The GCP Credential object or credential id

        Returns
        -------
        BatchPredictionJob
            Instance of BatchPredictonJob
        """
        intake_settings = {
            'type': 'gcp',
        }

        if credential is not None:
            if isinstance(credential, Credential):
                intake_settings['credential_id'] = credential.credential_id
            else:
                intake_settings['credential_id'] = credential

        output_settings = intake_settings.copy()

        intake_settings['url'] = source_url
        output_settings['url'] = destination_url

        kwargs.update({
            'intake_settings': intake_settings,
            'output_settings': output_settings,
        })

        return cls.score(deployment, **kwargs)

    @classmethod
    def score_from_existing(cls, batch_prediction_job_id):
        """ Create a new batch prediction job based on the settings from a previously created one

        Attributes
        ----------
        batch_prediction_job_id: str
            ID of the previous batch prediction job

        Returns
        -------
        BatchPredictionJob
            Instance of BatchPredictonJob
        """

        batch_job = super(BatchPredictionJob, cls).get(None, batch_prediction_job_id)

        job_data = {'predictionJobId': batch_job.id}

        response = cls._client.post(
            url=cls._from_existing_path(),
            json=job_data,
        )

        job_id = get_id_from_response(response)
        return BatchPredictionJob.get(job_id)

    @classmethod
    def get(cls, batch_prediction_job_id):
        """ Get batch prediction job

        Attributes
        ----------
        batch_prediction_job_id: str
            ID of batch prediction job

        Returns
        -------
        BatchPredictionJob
            Instance of BatchPredictonJob
        """
        batch_job = super(BatchPredictionJob, cls).get(
            None, batch_prediction_job_id)
        batch_job.id = batch_prediction_job_id

        return batch_job

    def download(self, fileobj):
        """ Downloads the CSV result of a prediction job

        Attributes
        ----------
        fileobj: file-like object
            Write CSV data to this file-like object

        """
        status = None
        for n in range(120):
            status = self.get_status()

            if status['status'] == 'ABORTED':
                raise RuntimeError('Job %s was aborted: %s' % (
                    self.id,
                    status['status_details'],
                ))

            if 'download' in status['links']:
                break
            time.sleep(1)

        if 'download' not in status['links']:
            raise RuntimeError('Could not detect download URL for job %s' % (
                self.id
            ))

        download_iter = self._client.get(
            status['links']['download'],
            stream=True,
        ).iter_content(chunk_size=8192)

        for chunk in download_iter:
            if chunk:
                fileobj.write(chunk)

    def delete(self):
        """
        Cancel this job. If this job has not finished running, it will be
        removed and canceled.
        """
        status = self.get_status()

        prediction_job_id = status['links']['self'].split('/')[-2]
        self._client.delete(self._job_path(None, prediction_job_id))

    def get_status(self):
        """Get status of batch prediction job

        Returns
        -------
        BatchPredictionJob status data
            Dict with job status
        """

        batch_job = super(BatchPredictionJob, self).get(None, self.id)
        batch_job.id = self.id

        return batch_job._safe_data

    @classmethod
    def list_by_status(cls, statuses=None):
        """Get jobs collection for specific set of statuses

        Attributes
        ----------
        statuses
            List of statuses to filter jobs ([ABORTED|COMPLETED...])
            if statuses is not provided, returns all jobs for user

        Returns
        -------
        BatchPredictionJob statuses
            List of job statses dicts with specific statuses
        """
        if statuses is not None:
            params = {'status': statuses}
        else:
            params = {}

        return list(
            pagination.unpaginate(cls._jobs_path(), params, cls._client))

    def _make_result_from_location(self, location, **kwargs):
        server_data = self._client.get(location)
        if server_data.status_code == 200:
            return server_data.content

    @classmethod
    def _data_and_completed_url_for_job(cls, url):
        response = cls._client.get(url, allow_redirects=False)

        if response.status_code == 200:
            data = response.json()
            if 'download' in data['links']:
                completed_url = data['links']['download']
            else:
                completed_url = data['links']['self']
            return data, completed_url
        else:
            e_msg = 'Server unexpectedly returned status code {}'
            raise errors.AsyncFailureError(
                e_msg.format(response.status_code))

    def __repr__(self):
        return 'BatchPredictionJob({}, \'{}\', status={})'.format(
            self.job_type, self.id, self.status
        )



import json

import dateutil
import pytest
import responses

from datarobot import Deployment
from datarobot.models.deployment import TargetDrift, FeatureDrift


class TestTargetDrift(object):
    @pytest.fixture
    def response_data(self, deployment_data):
        return {
            'model_id': deployment_data['model']['id'],
            'period': {
                'start': '2020-04-17T19:00:00.000000Z',
                'end': '2020-04-24T19:00:00.000000Z',
            },
            'metric': 'psi',
            'target_name': 'readmitted',
            'drift_score': 0.0040851491404388825,
            'sample_size': 5333,
            'baseline_sample_size': 504507,
        }

    @pytest.fixture
    def get_response(self, unittest_endpoint, deployment_data, response_data):
        url = '{}/deployments/{}/targetDrift/'.format(
            unittest_endpoint, deployment_data['id']
        )
        responses.add(
            responses.GET,
            url,
            status=200,
            content_type='application/json',
            body=json.dumps(response_data),
        )

    @responses.activate
    @pytest.mark.usefixtures('deployment_get_response', 'get_response')
    @pytest.mark.parametrize('from_deployment', [True, False])
    def test_target_drift(self, deployment_data, response_data, from_deployment):
        if from_deployment:
            target_drift = Deployment.get(deployment_data['id']).get_target_drift()
        else:
            target_drift = TargetDrift.get(deployment_data['id'])

        expected_start_date = dateutil.parser.parse(response_data['period']['start'])
        expected_end_date = dateutil.parser.parse(response_data['period']['end'])
        assert target_drift.period['start'] == expected_start_date
        assert target_drift.period['end'] == expected_end_date

        assert target_drift.model_id == response_data['model_id']
        assert target_drift.metric == response_data['metric']
        assert target_drift.target_name == response_data['target_name']
        assert target_drift.drift_score == response_data['drift_score']
        assert target_drift.sample_size == response_data['sample_size']
        assert target_drift.baseline_sample_size == response_data['baseline_sample_size']


class TestFeatureDrift(object):
    @pytest.fixture
    def response_data_first_page(self, deployment_data, unittest_endpoint):
        return {
            'model_id': deployment_data['model']['id'],
            'metric': 'psi',
            'period': {
                'start': '2020-04-20T19:00:00.000000Z',
                'end': '2020-04-27T19:00:00.000000Z',
            },
            'count': 2,
            'next': '{}/deployments/{}/featureDrift/?offset=2&limit=2&metric=psi'.format(
                unittest_endpoint, deployment_data['id']
            ),
            'previous': None,
            'data': [
                {
                    'featureImpact': 1.0,
                    'sampleSize': 157,
                    'name': 'age',
                    'baselineSampleSize': 2018030,
                    'driftScore': 4.169815947131828,
                },
                {
                    'featureImpact': 0.751510387892,
                    'sampleSize': 120,
                    'name': 'diag_1',
                    'baselineSampleSize': 2018030,
                    'driftScore': 1.2364572584578573,
                },
            ],
        }

    @pytest.fixture
    def response_data_second_page(self, deployment_data, unittest_endpoint):
        return {
            'model_id': deployment_data['model']['id'],
            'metric': 'psi',
            'period': {
                'start': '2020-04-20T19:00:00.000000Z',
                'end': '2020-04-27T19:00:00.000000Z',
            },
            'count': 1,
            'next': None,
            'previous': None,
            'data': [
                {
                    'featureImpact': 0.609462873473389,
                    'sampleSize': 147,
                    'name': 'gender',
                    'baselineSampleSize': 2018030,
                    'driftScore': 3.3003671706815156,
                },
            ],
        }

    @pytest.fixture
    def get_response(
        self,
        unittest_endpoint,
        deployment_data,
        response_data_first_page,
        response_data_second_page,
    ):
        url = '{}/deployments/{}/featureDrift/'.format(
            unittest_endpoint, deployment_data['id']
        )
        responses.add(
            responses.GET,
            url,
            status=200,
            content_type='application/json',
            body=json.dumps(response_data_first_page),
        )
        responses.add(
            responses.GET,
            response_data_first_page['next'],
            status=200,
            content_type='application/json',
            body=json.dumps(response_data_second_page),
        )

    @responses.activate
    @pytest.mark.usefixtures('deployment_get_response', 'get_response')
    @pytest.mark.parametrize('from_deployment', [True, False])
    def test_feature_drift(
        self,
        deployment_data,
        response_data_first_page,
        response_data_second_page,
        from_deployment,
    ):
        if from_deployment:
            feature_drifts = Deployment.get(deployment_data['id']).get_feature_drift()
        else:
            feature_drifts = FeatureDrift.list(deployment_data['id'])

        all_features = (
            response_data_first_page['data'] + response_data_second_page['data']
        )
        expected_start_date = dateutil.parser.parse(response_data_first_page['period']['start'])
        expected_end_date = dateutil.parser.parse(response_data_first_page['period']['end'])
        for index, feature_drift in enumerate(feature_drifts):
            assert feature_drift.period['start'] == expected_start_date
            assert feature_drift.period['end'] == expected_end_date

            assert feature_drift.model_id == response_data_first_page['model_id']
            assert feature_drift.metric == response_data_first_page['metric']
            assert feature_drift.name == all_features[index]['name']
            assert feature_drift.drift_score == all_features[index]['driftScore']
            assert feature_drift.feature_impact == all_features[index]['featureImpact']
            assert feature_drift.sample_size == all_features[index]['sampleSize']
            assert feature_drift.baseline_sample_size == all_features[index]['baselineSampleSize']



import pytest
import responses

from datarobot import Credential
from datarobot.utils import parse_time


@pytest.fixture
def credentials_server_resp():
    return {
        u'count': 2,
        u'data': [
            {
                u'creationDate': u'2020-02-03T10:13:27.411000Z',
                u'credentialId': u'5e37f247cf8a5f73842f6eab',
                u'credentialType': u's3',
                u'description': u'',
                u'name': u'my_s3_cred_1'
            },
            {
                u'creationDate': u'2020-02-10T14:52:07.258000Z',
                u'credentialId': u'5e416e17cf8a5f6e7afcc52f',
                u'credentialType': u'basic',
                u'description': u'???',
                u'name': u'my_basic_cred_1'
            },
        ],
        u'next': None,
        u'previous': None,
        u'totalCount': 2,
    }


@pytest.fixture
def s3_credential_server_resp(credentials_server_resp):
    return credentials_server_resp['data'][0]


@pytest.fixture
def basic_credential_server_resp(credentials_server_resp):
    return credentials_server_resp['data'][1]


@responses.activate
def test_list(credentials_endpoint, credentials_server_resp):
    responses.add(responses.GET, credentials_endpoint, json=credentials_server_resp)

    creds = Credential.list()

    for cred, server_payload in zip(creds, credentials_server_resp['data']):
        assert cred.credential_id == server_payload['credentialId']
        assert cred.name == server_payload['name']
        assert cred.credential_type == server_payload['credentialType']
        assert cred.creation_date == parse_time(server_payload['creationDate'])
        assert cred.description == server_payload['description']


@responses.activate
def test_get(credentials_endpoint, s3_credential_server_resp):
    credential_id = s3_credential_server_resp['credentialId']
    responses.add(
        responses.GET,
        '{}{}/'.format(credentials_endpoint, credential_id),
        json=s3_credential_server_resp
    )

    cred = Credential.get(credential_id)

    assert cred.credential_id == s3_credential_server_resp['credentialId']
    assert cred.name == s3_credential_server_resp['name']
    assert cred.credential_type == s3_credential_server_resp['credentialType']
    assert cred.creation_date == parse_time(s3_credential_server_resp['creationDate'])
    assert cred.description == s3_credential_server_resp['description']


@responses.activate
def test_create_basic(credentials_endpoint, basic_credential_server_resp):
    responses.add(responses.POST, credentials_endpoint, json=basic_credential_server_resp)

    cred = Credential.create_basic(
        name=basic_credential_server_resp['name'],
        user='USERNAME',
        password='PASSWORD',
        description=basic_credential_server_resp['description'],
    )

    assert cred.credential_id == basic_credential_server_resp['credentialId']
    assert cred.name == basic_credential_server_resp['name']
    assert cred.credential_type == basic_credential_server_resp['credentialType']
    assert cred.creation_date == parse_time(basic_credential_server_resp['creationDate'])
    assert cred.description == basic_credential_server_resp['description']


@responses.activate
def test_create_s3(credentials_endpoint, s3_credential_server_resp):
    responses.add(responses.POST, credentials_endpoint, json=s3_credential_server_resp)

    cred = Credential.create_s3(
        name=s3_credential_server_resp['name'],
        aws_access_key_id='?',
        aws_secret_access_key='!',
        description=s3_credential_server_resp['description'],
    )

    assert cred.credential_id == s3_credential_server_resp['credentialId']
    assert cred.name == s3_credential_server_resp['name']
    assert cred.credential_type == s3_credential_server_resp['credentialType']
    assert cred.creation_date == parse_time(s3_credential_server_resp['creationDate'])
    assert cred.description == s3_credential_server_resp['description']


@responses.activate
def test_delete(credentials_endpoint, s3_credential_server_resp):
    credential_id = s3_credential_server_resp['credentialId']
    responses.add(
        responses.GET,
        '{}{}/'.format(credentials_endpoint, credential_id),
        json=s3_credential_server_resp
    )
    responses.add(
        responses.DELETE,
        '{}{}/'.format(credentials_endpoint, credential_id),
    )

    cred = Credential.get(credential_id)
    cred.delete()



import pytest
import responses
from mock import patch

from datarobot.models import ExternalConfusionChart
from datarobot.errors import ClientError
from datarobot.utils import from_api


class TestExternalConfusionChart(object):

    @pytest.fixture
    def confusion_chart_classes(self):
        return

    @staticmethod
    def create_response_data(confusion_chart_data, count):
        confusion_chart_data['classes'] = ['1', '2', '3']
        return [
            {
                'data': confusion_chart_data,
                'datasetId': 'dataset_id_{}'.format(i),
            }
            for i in range(count)
        ]

    @pytest.fixture
    def url(self, project_id, model_id):
        return 'https://host_name.com/projects/{}/models/{}/datasetConfusionCharts/'.format(
            project_id,
            model_id
        )

    @pytest.fixture
    def metadata_url(self, project_id, model_id, dataset_id):
        return 'https://host_name.com/projects/{}/models/{}/datasetConfusionCharts/{}/metadata'\
            .format(project_id, model_id, dataset_id)

    @staticmethod
    def assert_equal_confusion_chart(confusion_chart_obj, api_chart_dict):
        one_chart_data = from_api(api_chart_dict['data'])
        assert confusion_chart_obj.dataset_id == api_chart_dict['datasetId']
        assert confusion_chart_obj.class_metrics == one_chart_data['class_metrics']
        assert confusion_chart_obj.confusion_matrix == one_chart_data['confusion_matrix']
        assert confusion_chart_obj.classes == one_chart_data['classes']

    def test_instantiation(self, confusion_chart_data):
        response_chart = self.create_response_data(confusion_chart_data, 1)[0]
        chart_obj = ExternalConfusionChart.from_server_data(response_chart)
        self.assert_equal_confusion_chart(chart_obj, response_chart)

    def test_future_proof(self, confusion_chart_data):
        response_chart = self.create_response_data(confusion_chart_data, 1)[0]
        data_with_future_keys = dict(response_chart, new_key='some future lift data')
        data_with_future_keys['data']['new_key'] = {'new_key': 'future class metrics'}
        ExternalConfusionChart.from_server_data(data_with_future_keys)

    @responses.activate
    @patch('datarobot.models.external_dataset_scores_insights.'
           'external_confusion_chart.DEFAULT_BATCH_SIZE', 10)
    def test_list_confusion_chart_all_results(self, confusion_chart_data, url, project_id,
                                              model_id):
        url_page1 = url + '?limit=10&offset=0'
        url_page2 = url + '?limit=10&offset=10'
        all_data = self.create_response_data(confusion_chart_data, 11)
        resp_data_1 = dict(count=10, previous=None, next=url_page2, data=all_data[:10])
        resp_data_2 = dict(count=1, previous=url_page1, next=None, data=all_data[10:])
        responses.add(responses.GET, url_page1, json=resp_data_1, status=200,
                      match_querystring=True)
        responses.add(responses.GET, url_page2, json=resp_data_2, status=200,
                      match_querystring=True)

        confusion_charts = ExternalConfusionChart.list(project_id, model_id, limit=0)
        assert len(confusion_charts) == 11
        for i, chart in enumerate(confusion_charts):
            assert isinstance(chart, ExternalConfusionChart)
            self.assert_equal_confusion_chart(chart, all_data[i])

    @responses.activate
    @patch('datarobot.models.external_dataset_scores_insights.'
           'external_confusion_chart.DEFAULT_BATCH_SIZE', 10)
    def test_list_confusion_chart_all_results_with_offset(self, confusion_chart_data, url,
                                                          project_id, model_id):
        url_page1 = url + '?limit=10&offset=2'
        url_page2 = url + '?limit=10&offset=12'
        all_data = self.create_response_data(confusion_chart_data, 15)
        resp_data_1 = dict(count=10, previous=None, next=url_page2, data=all_data[2:12])
        resp_data_2 = dict(count=3, previous=url_page1, next=None, data=all_data[12:])
        responses.add(responses.GET, url_page1, json=resp_data_1, status=200,
                      match_querystring=True)
        responses.add(responses.GET, url_page2, json=resp_data_2, status=200,
                      match_querystring=True)

        confusion_charts = ExternalConfusionChart.list(project_id, model_id, limit=0, offset=2)
        assert len(confusion_charts) == 13
        assert confusion_charts[0].dataset_id == 'dataset_id_2'

    @responses.activate
    def test_list_confusion_chart_limited(self, confusion_chart_data, url, project_id, model_id):
        url_page = url + '?limit=3&offset=2'
        all_data = self.create_response_data(confusion_chart_data, 5)
        resp_data = dict(count=3, previous=None, next=None, data=all_data[2:5])
        responses.add(responses.GET, url_page, json=resp_data, status=200,
                      match_querystring=True)
        confusion_charts = ExternalConfusionChart.list(project_id, model_id, limit=3, offset=2)
        assert len(confusion_charts) == 3
        for i, chart in enumerate(confusion_charts):
            assert isinstance(chart, ExternalConfusionChart)
            self.assert_equal_confusion_chart(chart, all_data[i + 2])

    @responses.activate
    def test_get_confusion_chart_success(self, confusion_chart_data, url, project_id, model_id):
        dataset_id = 'dataset_id_0'
        new_classes = ['4', '5', '6']
        url_page = url + '{}/'.format(dataset_id)
        metadata_url = url_page + 'metadata/'
        all_data = self.create_response_data(confusion_chart_data, 2)
        metadata_response = {'classNames': new_classes}

        responses.add(responses.GET, metadata_url, json=metadata_response, status=200)
        responses.add(responses.GET, url_page, json=all_data[1], status=200,
                      match_querystring=True)

        chart = ExternalConfusionChart.get(project_id, model_id, dataset_id)
        expected_chart_dict = all_data[1]
        expected_chart_dict['data']['classes'] = ['4', '5', '6']
        assert isinstance(chart, ExternalConfusionChart)
        self.assert_equal_confusion_chart(chart, expected_chart_dict)

    @responses.activate
    def test_get_confusion_chart_failure(self, url, project_id, model_id):
        dataset_id = 'dataset_id_0'
        url_page = url + '{}/'.format(dataset_id)

        exception_text = 'No confusion chart for dataset {}'.format(dataset_id)

        responses.add(responses.GET, url_page, json={'message': exception_text}, status=404)
        with pytest.raises(ClientError, match=exception_text):
            ExternalConfusionChart.get(project_id, model_id, dataset_id)



import pytest
import json
import responses

from datarobot import errors
from tests.utils import request_body_to_json
from datarobot.models.secondary_dataset import (
    DatasetConfiguration,
    SecondaryDataset,
    SecondaryDatasetConfigurations,
)


@pytest.fixture
def ids():
    return {
        'pid': '5a530498d5c1f302d6d17699',
        'graph_id': '5a530498d5c1f302d6d17111',
        'identifier_1': 'test_dataset_1',
        'catalog_version_id_1': '5a530498d5c1f302d6d17121',
        'catalog_id_1': '5a530498d5c1f302d6d17115',
        'identifier_2': 'test_dataset_2',
        'catalog_version_id_2': '6a530498d5c1f302d6d17121',
        'catalog_id_2': '6a530498d5c1f302d6d17115',
    }


@pytest.fixture
def dataset_configurations(ids):
    secondary_dataset_1 = SecondaryDataset(
        identifier=ids['identifier_1'],
        catalog_version_id=ids['catalog_version_id_1'],
        catalog_id=ids['catalog_id_1'],
        snapshot_policy='specified',
    )
    secondary_dataset_2 = SecondaryDataset(
        identifier=ids['identifier_2'],
        catalog_version_id=ids['catalog_version_id_2'],
        catalog_id=ids['catalog_id_2'],
        snapshot_policy='specified',
    )
    return [
        DatasetConfiguration(
            feature_engineering_graph_id=ids['graph_id'],
            secondary_datasets=[secondary_dataset_1, secondary_dataset_2],
        )
    ]


@pytest.fixture
def secondary_dataset_configurations_response(ids):
    return {
        'projectId': ids['pid'],
        'config': [
            {
                'secondaryDatasets': [
                    {
                        'snapshotPolicy': 'specified',
                        'identifier': ids['identifier_1'],
                        'catalogVersionId': ids['catalog_version_id_1'],
                        'catalogId': ids['catalog_id_1'],
                    },
                    {
                        'snapshotPolicy': 'specified',
                        'identifier': ids['identifier_2'],
                        'catalogVersionId': ids['catalog_version_id_2'],
                        'catalogId': ids['catalog_id_2'],
                    },
                ],
                'featureEngineeringGraphId': ids['graph_id'],
            }
        ],
        'id': '5df109112ca582033ff44084',
    }


@responses.activate
def test_create_secondary_dataset_configurations(
    ids, dataset_configurations, secondary_dataset_configurations_response
):
    url = 'https://host_name.com/projects/{}/secondaryDatasetsConfigurations/'.format(
        ids['pid']
    )
    responses.add(
        responses.POST,
        url,
        status=201,
        body=json.dumps(secondary_dataset_configurations_response),
    )

    result = SecondaryDatasetConfigurations.create(ids['pid'], dataset_configurations)

    assert responses.calls[0].request.method == 'POST'

    # verify request payload
    payload = request_body_to_json(responses.calls[0].request)
    conf = payload['conf'][0]
    assert conf['featureEngineeringGraphId'] == ids['graph_id']
    dataset_confs = conf['secondaryDatasets']
    expected_conf_sets = [
        d.to_dict() for d in dataset_configurations[0].secondary_datasets
    ]
    assert len(dataset_confs) == len(expected_conf_sets)
    identifier = dataset_confs[0]['identifier']
    expected_dataset = [d for d in expected_conf_sets if d['identifier'] == identifier][
        0
    ]
    assert dataset_confs[0] == expected_dataset

    # verify response
    assert isinstance(result, SecondaryDatasetConfigurations)
    assert result.project_id == ids['pid']
    dataset_conf = result.config[0]
    assert isinstance(dataset_conf, DatasetConfiguration)
    assert dataset_conf.feature_engineering_graph_id == ids['graph_id']
    dataset = dataset_conf.secondary_datasets[0]
    assert isinstance(dataset, SecondaryDataset)
    identifier = dataset.identifier
    expected_dataset = [d for d in expected_conf_sets if d['identifier'] == identifier][
        0
    ]
    assert dataset.catalog_id == expected_dataset['catalogId']
    assert dataset.catalog_version_id == expected_dataset['catalogVersionId']
    assert dataset.snapshot_policy == expected_dataset['snapshotPolicy']


@responses.activate
def test_failure_in_create_secondary_dataset_configurations__status_error_code(
    ids, dataset_configurations
):
    url = 'https://host_name.com/projects/{}/secondaryDatasetsConfigurations/'.format(
        ids['pid']
    )
    responses.add(responses.POST, url, status=404, body='')

    with pytest.raises(errors.ClientError):
        SecondaryDatasetConfigurations.create(ids['pid'], dataset_configurations)


@responses.activate
def test_secondary_dataset_configurations_retrieve(
    ids, secondary_dataset_configurations_response,
):
    config_id = secondary_dataset_configurations_response['id']
    pid = secondary_dataset_configurations_response['projectId']
    dataset_configurations = secondary_dataset_configurations_response['config']
    url = 'https://host_name.com/projects/{}/secondaryDatasetsConfigurations/{}/'.format(
        pid, config_id
    )
    responses.add(responses.GET, url, json=secondary_dataset_configurations_response)

    secondary_dataset_config = SecondaryDatasetConfigurations(
        id=config_id, project_id=pid
    ).get()

    assert responses.calls[0].request.method == 'GET'
    assert secondary_dataset_config.id == config_id
    assert secondary_dataset_config.project_id == pid
    assert len(secondary_dataset_config.config) == 1

    dataset_conf = secondary_dataset_config.config[0]
    assert isinstance(dataset_conf, DatasetConfiguration)
    assert dataset_conf.feature_engineering_graph_id == ids['graph_id']
    dataset = dataset_conf.secondary_datasets[0]
    assert isinstance(dataset, SecondaryDataset)
    identifier = dataset.identifier
    expected_config = dataset_configurations[0]['secondaryDatasets']
    expected_dataset = [d for d in expected_config if d['identifier'] == identifier][0]
    assert dataset.catalog_id == expected_dataset['catalogId']
    assert dataset.catalog_version_id == expected_dataset['catalogVersionId']
    assert dataset.snapshot_policy == expected_dataset['snapshotPolicy']


@responses.activate
def test_secondary_dataset_configurations_deletion(
    secondary_dataset_configurations_response,
):
    config_id = secondary_dataset_configurations_response['id']
    pid = secondary_dataset_configurations_response['projectId']
    url = 'https://host_name.com/projects/{}/secondaryDatasetsConfigurations/{}/'.format(
        pid, config_id
    )
    responses.add(responses.GET, url, json=secondary_dataset_configurations_response)
    responses.add(responses.DELETE, url)
    secondary_dataset_config = SecondaryDatasetConfigurations(
        id=config_id, project_id=pid
    ).get()
    secondary_dataset_config.delete()
    assert responses.calls[1].request.method == responses.DELETE
    assert responses.calls[1].request.url == url



import trafaret as t

from datarobot.models.api_object import APIObject
from datarobot.utils import encode_utf8_if_py2

_PercentageFloat = t.Float(gte=0, lte=1)  # 0.0 <= x <= 1.0

_ClassPercentage = t.Dict({
    t.Key('other_class_name'): t.String,
    t.Key('percentage'): _PercentageFloat,
}).ignore_extra('*')

_ClassMetrics = t.Dict({
    t.Key('class_name'): t.String,
    # blue bars on the axis
    t.Key('actual_count'): t.Int,
    t.Key('predicted_count'): t.Int,
    # one vs all metrics
    t.Key('f1'): _PercentageFloat,
    t.Key('recall'): _PercentageFloat,
    t.Key('precision'): _PercentageFloat,
    t.Key('confusion_matrix_one_vs_all'): t.List(t.List(t.Int)),
    t.Key('was_actual_percentages'): t.List(_ClassPercentage),
    t.Key('was_predicted_percentages'): t.List(_ClassPercentage),
}).ignore_extra('*')

ConfusionChartTrafaret = t.Dict({
    # available classes
    t.Key('classes'): t.List(t.String),
    # NxN confusion matrix
    t.Key('confusion_matrix'): t.List(t.List(t.Int)),
    t.Key('class_metrics'): t.List(_ClassMetrics),
}).ignore_extra('*')


class ConfusionChart(APIObject):
    """ Confusion Chart data for model.

    Notes
    -----
    ``ClassMetrics`` is a dict containing the following:

        * ``class_name`` (string) name of the class
        * ``actual_count`` (int) number of times this class is seen in the validation data
        * ``predicted_count`` (int) number of times this class has been predicted for the \
          validation data
        * ``f1`` (float) F1 score
        * ``recall`` (float) recall score
        * ``precision`` (float) precision score
        * ``was_actual_percentages`` (list of dict) one vs all actual percentages in format \
          specified below.
            * ``other_class_name`` (string) the name of the other class
            * ``percentage`` (float) the percentage of the times this class was predicted when is \
              was actually class (from 0 to 1)
        * ``was_predicted_percentages`` (list of dict) one vs all predicted percentages in format \
          specified below.
            * ``other_class_name`` (string) the name of the other class
            * ``percentage`` (float) the percentage of the times this class was actual predicted \
              (from 0 to 1)
        * ``confusion_matrix_one_vs_all`` (list of list) 2d list representing 2x2 one vs all matrix.
            * This represents the True/False Negative/Positive rates as integer for each class. \
              The data structure looks like:
            * ``[ [ True Negative, False Positive ], [ False Negative, True Positive ] ]``


    Attributes
    ----------
    source : str
        Confusion Chart data source. Can be 'validation', 'crossValidation' or 'holdout'.
    raw_data : dict
        All of the raw data for the Confusion Chart
    confusion_matrix : list of list
        The NxN confusion matrix
    classes : list
        The names of each of the classes
    class_metrics : list of dicts
        List of dicts with schema described as ``ClassMetrics`` above.
    source_model_id : str
        ID of the model this Confusion chart represents; in some cases,
        insights from the parent of a frozen model may be used

    """

    ConfusionChartWrapper = t.Dict({
        t.Key('data'): ConfusionChartTrafaret,
        t.Key('source'): t.String,
        t.Key('source_model_id'): t.String
    }).ignore_extra('*')

    _converter = ConfusionChartWrapper

    def __init__(self, source, data, source_model_id):
        self.source = source
        self.raw_data = data
        self.class_metrics = data['class_metrics']
        self.confusion_matrix = data['confusion_matrix']
        self.classes = data['classes']
        self.source_model_id = source_model_id

    def __repr__(self):
        return encode_utf8_if_py2(u'ConfusionChart({})'.format(self.source))



import responses
import json
import pytest
from uuid import uuid4

from datarobot import PredictionServer


@pytest.fixture()
def prediction_server_data_cloud():
    return {
        'id': '5c2aad800000000000000000',
        'url': 'https://predictions.example.com',
        'datarobot-key': str(uuid4())
    }


@pytest.fixture()
def prediction_server_data_on_prem():
    return {
        'id': None,
        'url': 'https://predictions.example.com',
        'datarobot-key': None,
    }


def make_prediction_server_list_response(prediction_server_data):
    return {
        'count': 1,
        'next': None,
        'previous': None,
        'data': [prediction_server_data]
    }


@responses.activate
@pytest.mark.parametrize(
    'prediction_server_data_fixture',
    ['prediction_server_data_cloud', 'prediction_server_data_on_prem'],
)
def test_prediction_server_list(request, prediction_server_data_fixture):
    prediction_server_data = request.getfixturevalue(prediction_server_data_fixture)

    server_list_response = make_prediction_server_list_response(prediction_server_data)
    responses.add(
        responses.GET,
        'https://host_name.com/predictionServers/',
        status=200,
        content_type='application/json',
        body=json.dumps(server_list_response)
    )

    prediction_servers = PredictionServer.list()
    assert len(prediction_servers) == 1

    prediction_server = prediction_servers[0]
    assert prediction_server.id == prediction_server_data['id']
    assert prediction_server.url == prediction_server_data['url']
    assert prediction_server.datarobot_key == prediction_server_data['datarobot-key']

    repr = 'PredictionServer({url})'.format(url=prediction_server_data['url'])
    assert str(prediction_server) == repr



import json
from six.moves.urllib.parse import urljoin

from datarobot import enums

import pytest
import responses


@pytest.fixture
def model_url(project_url, model_id):
    return urljoin(project_url, "models/{}/".format(model_id))


@pytest.fixture
def feature_impact_url(project_id, model_id):
    return 'https://host_name.com/projects/{}/models/{}/featureImpact/'.format(project_id, model_id)


@pytest.fixture
def multiclass_feature_impact_url(project_id, model_id):
    return 'https://host_name.com/projects/{}/models/{}/multiclassFeatureImpact/'.format(
        project_id, model_id)


@pytest.fixture
def feature_impact_job_creation_response(feature_impact_url, job_url):
    responses.add(
        responses.POST,
        feature_impact_url,
        body='',
        status=202,
        adding_headers={'Location': job_url})


@pytest.fixture
def multiclass_feature_impact_server_data():
    return {
        'ranRedundancyDetection': False,
        'classFeatureImpacts': [
            {'featureImpacts': [
                {'featureName': 'Petal Width',
                 'redundantWith': None,
                 'impactUnnormalized': 0.157,
                 'impactNormalized': 0.701},
                {'featureName': 'Petal Length',
                 'redundantWith': None,
                 'impactUnnormalized': 0.121,
                 'impactNormalized': 0.539},
                {'featureName': 'Sepal Width',
                 'redundantWith': None,
                 'impactUnnormalized': 0.225,
                 'impactNormalized': 1.0},
                {'featureName': 'Sepal Length',
                 'redundantWith': None,
                 'impactUnnormalized': 0.057,
                 'impactNormalized': 0.257}],
                'class': 'setosa'},
            {'featureImpacts': [
                {'featureName': 'Petal Width',
                 'redundantWith': None,
                 'impactUnnormalized': 0.302,
                 'impactNormalized': 1.0},
                {'featureName': 'Petal Length',
                 'redundantWith': None,
                 'impactUnnormalized': 0.229,
                 'impactNormalized': 0.758},
                {'featureName': 'Sepal Width',
                 'redundantWith': None,
                 'impactUnnormalized': 0.221,
                 'impactNormalized': 0.730},
                {'featureName': 'Sepal Length',
                 'redundantWith': None,
                 'impactUnnormalized': 0.073,
                 'impactNormalized': 0.241}],
                'class': 'versicolor'},
            {'featureImpacts': [
                {'featureName': 'Petal Width',
                 'redundantWith': None,
                 'impactUnnormalized': 0.273,
                 'impactNormalized': 1.0},
                {'featureName': 'Petal Length',
                 'redundantWith': None,
                 'impactUnnormalized': 0.174,
                 'impactNormalized': 0.640},
                {'featureName': 'Sepal Width',
                 'redundantWith': None,
                 'impactUnnormalized': -0.001,
                 'impactNormalized': -0.006},
                {'featureName': 'Sepal Length',
                 'redundantWith': None,
                 'impactUnnormalized': 0.018,
                 'impactNormalized': 0.067}],
                'class': 'virginica'}
        ]
    }


@pytest.fixture
def feature_impact_response(feature_impact_server_data, feature_impact_url):
    body = json.dumps(feature_impact_server_data)
    responses.add(
        responses.GET,
        feature_impact_url,
        status=200,
        content_type='application/json',
        body=body
    )


@pytest.fixture
def multiclass_feature_impact_response(multiclass_feature_impact_server_data,
                                       multiclass_feature_impact_url):
    body = json.dumps(multiclass_feature_impact_server_data)
    responses.add(
        responses.GET,
        multiclass_feature_impact_url,
        status=200,
        content_type='application/json',
        body=body
    )


@pytest.fixture
def feature_impact_job_running_server_data(base_job_running_server_data):
    return dict(base_job_running_server_data, jobType='featureImpact')


@pytest.fixture
def feature_impact_job_finished_server_data(base_job_completed_server_data):
    return dict(base_job_completed_server_data, jobType='featureImpact')


@pytest.fixture
def approximate_job_running_server_data(base_job_running_server_data):
    return dict(base_job_running_server_data, jobType=enums.JOB_TYPE.PRIME_RULESETS)


@pytest.fixture
def approximate_job_finished_server_data(base_job_completed_server_data):
    return dict(base_job_completed_server_data, jobType=enums.JOB_TYPE.PRIME_RULESETS)


@pytest.fixture
def prime_validation_job_running_server_data(base_job_running_server_data):
    return dict(base_job_running_server_data, jobType=enums.JOB_TYPE.PRIME_VALIDATION)


@pytest.fixture
def prime_validation_job_finished_server_data(base_job_completed_server_data):
    return dict(base_job_completed_server_data, jobType=enums.JOB_TYPE.PRIME_VALIDATION)


@pytest.fixture
def feature_impact_completed_response(feature_impact_job_finished_server_data,
                                      job_url, feature_impact_url):
    """
    Loads a response that the given job is a featureImpact job, and is in completed
    """
    responses.add(responses.GET,
                  job_url,
                  body=json.dumps(feature_impact_job_finished_server_data),
                  status=303,
                  adding_headers={'Location': feature_impact_url},
                  content_type='application/json')


@pytest.fixture
def feature_impact_previously_ran_response(project_url, model_id, job_id):
    """
    Loads a response that the given model has already ran its feature impact
     """
    body = {
        'message': 'Feature Impact is in progress for this model.',
        'errorName': 'JobAlreadyAdded',
        'jobId': job_id
        }
    responses.add(responses.POST,
                  '{}models/{}/featureImpact/'.format(project_url, model_id),
                  body=json.dumps(body),
                  status=422,
                  content_type='application/json')


@pytest.fixture
def feature_impact_running_response(feature_impact_job_running_server_data, job_url):
    """
    Loads a response that the given job is a featureImpact job, and is running
    """
    responses.add(responses.GET,
                  job_url,
                  body=json.dumps(feature_impact_job_running_server_data),
                  status=200,
                  content_type='application/json')


@pytest.fixture
def approximate_job_creation_response(project_url, model_id, job_url):
    responses.add(responses.POST,
                  '{}models/{}/primeRulesets/'.format(project_url, model_id),
                  body='',
                  status=202,
                  adding_headers={'Location': job_url})


@pytest.fixture
def approximate_completed_response(approximate_job_finished_server_data, job_url,
                                   project_url, model_id,
                                   ruleset_without_model_server_data,
                                   ruleset_with_model_server_data):
    rulesets_url = '{}models/{}/primeRulesets/'.format(project_url, model_id)
    responses.add(responses.GET, job_url, body=json.dumps(approximate_job_finished_server_data),
                  status=303,
                  adding_headers={'Location': rulesets_url},
                  content_type='application/json')
    responses.add(responses.GET, rulesets_url,
                  body=json.dumps([ruleset_with_model_server_data,
                                   ruleset_without_model_server_data]),
                  content_type='application/json')


@pytest.fixture
def prime_validation_job_creation_response(project_url, job_url):
    responses.add(responses.POST, '{}primeFiles/'.format(project_url), body='', status=202,
                  adding_headers={'Location': job_url})


@pytest.fixture
def prime_validation_job_completed_response(prime_validation_job_finished_server_data,
                                            job_url, project_url, prime_file_server_data):
    file_url = '{}primeFiles/{}/'.format(project_url, prime_file_server_data['id'])
    responses.add(responses.GET, job_url,
                  body=json.dumps(prime_validation_job_finished_server_data),
                  status=303, adding_headers={'Location': file_url},
                  content_type='application/json')
    responses.add(responses.GET, file_url, body=json.dumps(prime_file_server_data),
                  content_type='application/json')


@pytest.fixture
def mocked_execution_environment_versions():
    return {
        "count": 3,
        "totalCount": 3,
        "next": None,
        "previous": None,
        "data": [
            {
                "id": "5cf4d3f5f930e26daac18a1a",
                "environmentId": "5cf4d3f5f930e26daac18a1xx",
                "label": "Version 1",
                "description": "",
                "buildStatus": "success",
                "created": "2019-09-28T15:19:26.587583Z",
            },
            {
                "id": "5cf4d3f5f930e26daac18a2a",
                "environmentId": "5cf4d3f5f930e26daac18a1xx",
                "label": "Version 1",
                "description": "",
                "buildStatus": "processing",
                "created": "2019-09-28T15:19:26.587583Z",
            },
            {
                "id": "5cf4d3f5f930e26daac18a3a",
                "environmentId": "5cf4d3f5f930e26daac18a1xx",
                "label": "Version 1",
                "description": "some description",
                "buildStatus": "failed",
                "created": "2019-09-28T15:19:26.587583Z",
            },
        ],
    }


@pytest.fixture
def mocked_execution_environment_version(mocked_execution_environment_versions):
    return mocked_execution_environment_versions["data"][0]


@pytest.fixture
def mocked_execution_environments(mocked_execution_environment_version):
    return {
        "count": 2,
        "totalCount": 2,
        "next": None,
        "previous": None,
        "data": [
            {
                "id": "5cf4d3f5f930e26daac18a1a",
                "name": "env1",
                "description": "some description",
                "programmingLanguage": "python",
                "isPublic": True,
                "created": "2019-09-28T15:19:26.587583Z",
                "latestVersion": mocked_execution_environment_version,
            },
            {
                "id": "5cf4d3f5f930e26daac18a1a",
                "name": "env2",
                "description": "",
                "programmingLanguage": "other",
                "isPublic": False,
                "created": "2019-09-28T15:19:26.587583Z",
                "latestVersion": None,
            },
        ],
    }


@pytest.fixture
def mocked_execution_environment(mocked_execution_environments):
    return mocked_execution_environments["data"][0]


@pytest.fixture
def mocked_versions():
    return {
        "count": 2,
        "totalCount": 2,
        "next": None,
        "previous": None,
        "data": [
            {
                "id": "5cf4d3f5f930e26daac18a1a",
                "customModelId": "5cf4d3f5f930e26daac18a1xx",
                "label": "Version 1",
                "description": "",
                "versionMinor": 0,
                "versionMajor": 1,
                "isFrozen": False,
                "items": [
                    {
                        "id": "5cf4d3f5f930e26daac18xxx",
                        "fileName": "name",
                        "filePath": "path",
                        "fileSource": "source",
                        "created": "2019-09-28T15:19:26.587583Z",
                    }
                ],
                "created": "2019-09-28T15:19:26.587583Z",
            },
            {
                "id": "5cf4d3f5f930e26daac18aBB",
                "customModelId": "5cf4d3f5f930e26daac18a1BB",
                "label": "Version 2",
                "description": "ss",
                "versionMinor": 2,
                "versionMajor": 3,
                "isFrozen": True,
                "items": [],
                "created": "2019-09-28T15:19:26.587583Z",
            },
        ],
    }


@pytest.fixture
def mocked_custom_model_version(mocked_versions):
    return mocked_versions['data'][0]


@pytest.fixture
def mocked_custom_model_version2(mocked_versions):
    return mocked_versions['data'][1]


@pytest.fixture
def shap_impact_url(project_id, model_id):
    return 'https://host_name.com/projects/{}/models/{}/shapImpact/'.format(project_id, model_id)


@pytest.fixture
def shap_impact_job_creation_response(shap_impact_url, job_url):
    responses.add(
        responses.POST,
        shap_impact_url,
        body='',
        status=202,
        adding_headers={'Location': job_url}
    )


@pytest.fixture
def shap_impact_response(shap_impact_server_data, shap_impact_url):
    body = json.dumps(shap_impact_server_data)
    responses.add(
        responses.GET,
        shap_impact_url,
        status=200,
        content_type='application/json',
        body=body
    )


@pytest.fixture
def shap_impact_job_running_server_data(base_job_running_server_data):
    return dict(base_job_running_server_data, jobType='shapImpact')


@pytest.fixture
def shap_impact_job_finished_server_data(base_job_completed_server_data):
    return dict(base_job_completed_server_data, jobType='shapImpact')


@pytest.fixture
def shap_impact_running_response(shap_impact_job_running_server_data, job_url):
    responses.add(
        responses.GET,
        job_url,
        body=json.dumps(shap_impact_job_running_server_data),
        status=200,
        content_type='application/json'
    )


@pytest.fixture
def shap_impact_completed_response(shap_impact_job_finished_server_data, job_url, shap_impact_url):
    responses.add(
        responses.GET,
        job_url,
        body=json.dumps(shap_impact_job_finished_server_data),
        status=303,
        adding_headers={'Location': shap_impact_url},
        content_type='application/json'
    )


@pytest.fixture
def shap_impact_previously_ran_response(project_url, model_id, job_id, job_url):
    body = {
        'message': 'SHAP impact already exists.',
        'errorName': 'JobAlreadyAdded',
        'jobId': job_id
        }
    responses.add(
        responses.POST,
        '{}models/{}/shapImpact/'.format(project_url, model_id),
        body=json.dumps(body),
        status=202,
        content_type='application/json',
        adding_headers={'Location': job_url}
    )



import json

import pytest
import responses

from datarobot import Deployment
from datarobot.utils import from_api


class TestListFeatures(object):
    @pytest.fixture
    def response_data(self, deployment_data):
        return {
            'count': 40,
            'next': None,
            'previous': None,
            'data': [
                {
                    'importance': 0.12,
                    'featureType': 'Categorical',
                    'knownInAdvance': False,
                    'name': 'Feature1'
                }, {
                    'importance': 0.05,
                    'featureType': 'Numeric',
                    'knownInAdvance': True,
                    'name': 'Feature2'
                }
            ]
        }

    @pytest.fixture
    def response(self, unittest_endpoint, deployment_data, response_data):
        url = '{}/deployments/{}/features/'.format(unittest_endpoint, deployment_data['id'])
        responses.add(
            responses.GET,
            url,
            status=200,
            content_type='application/json',
            body=json.dumps(response_data)
        )

    @responses.activate
    @pytest.mark.usefixtures('deployment_get_response', 'response')
    def test_retrieve(self, deployment_data, response_data):
        actual_features = Deployment.get(deployment_data['id']).get_features()
        expected_features = from_api(response_data['data'], keep_null_keys=True)
        assert actual_features == expected_features



from __future__ import division

import math
import random
from datetime import datetime

import pandas as pd
import pytest
import responses

from datarobot import Deployment
from tests.utils import request_body_to_json


class TestSubmitActuals(object):
    @pytest.fixture()
    def response(self, unittest_endpoint, deployment_data):
        url = '{}/{}/{}/actuals/fromJSON/'.format(
            unittest_endpoint, 'deployments', deployment_data['id']
        )
        status_url = '{}/STATUS_ID/'.format(unittest_endpoint)
        accuracy_url = '{}/{}/{}/accuracy/'.format(
            unittest_endpoint, 'deployments', deployment_data['id']
        )
        responses.add(responses.POST, url, status=202, headers={'Location': status_url})
        responses.add(
            responses.GET, status_url, status=303, headers={'Location': accuracy_url}
        )

    @responses.activate
    @pytest.mark.usefixtures('deployment_get_response')
    @pytest.mark.parametrize('actuals_data', [None, [], 'abc', pd.DataFrame()])
    def test_bad_data(self, deployment_data, actuals_data):
        """When submitting actuals with bad data, `ValueError` is raised."""
        with pytest.raises(ValueError):
            deployment = Deployment.get(deployment_data['id'])
            deployment.submit_actuals(actuals_data)

    @responses.activate
    @pytest.mark.usefixtures('deployment_get_response')
    @pytest.mark.parametrize('batch_size', [None, 3.7, 0, -3])
    def test_bad_batch_size(self, deployment_data, batch_size):
        """When submitting actuals with bad batch_size, `ValueError` is raised."""
        with pytest.raises(ValueError):
            deployment = Deployment.get(deployment_data['id'])
            deployment.submit_actuals(
                [{'association_id': 'abc', 'actual_value': 'True'}],
                batch_size=batch_size,
            )

    @responses.activate
    @pytest.mark.usefixtures('deployment_get_response', 'response')
    @pytest.mark.parametrize(
        'actuals_data, expected_payload',
        [
            (
                [{'association_id': 'abc', 'actual_value': 'True'}],
                {'data': [{'associationId': 'abc', 'actualValue': 'True'}]},
            ),
            (
                [{'association_id': 'abc', 'actual_value': 'True'}],
                {'data': [{'associationId': 'abc', 'actualValue': 'True'}]},
            ),
            (
                [
                    {
                        'association_id': 'abc',
                        'actual_value': 'True',
                        'was_acted_on': True,
                    }
                ],
                {
                    'data': [
                        {
                            'associationId': 'abc',
                            'actualValue': 'True',
                            'wasActedOn': True,
                        }
                    ]
                },
            ),
            (
                [
                    {
                        'association_id': 'abc',
                        'actual_value': 'True',
                        'timestamp': datetime(2019, 8, 1, 12, 30),
                    }
                ],
                {
                    'data': [
                        {
                            'associationId': 'abc',
                            'actualValue': 'True',
                            'timestamp': '2019-08-01T12:30:00+00:00',
                        }
                    ]
                },
            ),
        ],
    )
    def test_submit(self, deployment_data, actuals_data, expected_payload):
        deployment = Deployment.get(deployment_data['id'])
        deployment.submit_actuals(actuals_data)
        actual_payload = request_body_to_json(responses.calls[1].request)
        assert actual_payload == expected_payload

    @responses.activate
    @pytest.mark.usefixtures('deployment_get_response', 'response')
    def test_submit_with_dataframe(self, deployment_data):
        actuals_data = pd.DataFrame(
            [
                ['a', 1, None, None],
                ['b', 2, True, None],
                ['c', 3, None, datetime(2019, 8, 1, 12, 30)],
                ['d', 4, False, datetime(2019, 8, 1, 12, 30)],
            ],
            columns=['association_id', 'actual_value', 'was_acted_on', 'timestamp'],
        )
        expected_payload = {
            'data': [
                {'associationId': 'a', 'actualValue': 1},
                {'associationId': 'b', 'actualValue': 2, 'wasActedOn': True},
                {
                    'associationId': 'c',
                    'actualValue': 3,
                    'timestamp': '2019-08-01T12:30:00+00:00',
                },
                {
                    'associationId': 'd',
                    'actualValue': 4,
                    'wasActedOn': False,
                    'timestamp': '2019-08-01T12:30:00+00:00',
                },
            ]
        }

        deployment = Deployment.get(deployment_data['id'])
        deployment.submit_actuals(actuals_data)
        actual_payload = request_body_to_json(responses.calls[1].request)
        assert actual_payload == expected_payload

    @responses.activate
    @pytest.mark.usefixtures('deployment_get_response', 'response')
    def test_batched_submit(self, deployment_data):
        """Test submit actuals in batches"""

        total_actuals_count = 536
        batch_size = 100

        # generate some test data
        random.seed(1005)
        actuals = [
            {
                'association_id': 'abc',
                'actual_value': random.randrange(0, 100),
                'was_acted_on': bool(random.getrandbits(1)),
            }
            for _ in range(total_actuals_count)
        ]
        deployment = Deployment.get(deployment_data['id'])
        deployment.submit_actuals(actuals, batch_size=batch_size)

        # submit actuals
        post_request_count = 0
        payload_data = []
        for call in responses.calls:
            if call.request.method == 'POST':
                post_request_count += 1
                payload = request_body_to_json(call.request)
                payload_data += payload['data']

        # test if API calls are made correctly
        assert post_request_count == math.ceil(total_actuals_count / batch_size)
        assert len(actuals) == len(payload_data)
        for actual, payload in zip(actuals, payload_data):
            assert actual['association_id'] == payload['associationId']
            assert actual['actual_value'] == payload['actualValue']
            assert actual['was_acted_on'] == payload['wasActedOn']



import json

import pytest
import responses

from datarobot import Deployment
from datarobot.utils import from_api
from tests.utils import request_body_to_json


@pytest.fixture()
def deployment_settings_response_data():
    return {
        'targetDrift': {'enabled': True},
        'featureDrift': {'enabled': True},
        'associationId': {
            'columnNames': ['column'],
            'requiredInPredictionRequests': False
        },
        'predictionWarning': {'enabled': True,
                              'customBoundaries': {
                                  'upper': 1337,
                                  'lower': 0
                              }},
        'predictionIntervals': {'enabled': True, 'percentiles': [80]},
        'predictionsDataCollection': {'enabled': True},
        'otherSettings': {'test': 'value'}
    }


@pytest.fixture()
def deployment_settings_get_response(unittest_endpoint, deployment_data,
                                     deployment_settings_response_data):
    deployment_id = deployment_data['id']
    url = '{}/deployments/{}/settings/'.format(unittest_endpoint, deployment_id)
    responses.add(
        responses.GET,
        url,
        body=json.dumps(deployment_settings_response_data),
        status=200,
        content_type='application/json')


@pytest.fixture()
def deployment_settings_update_response(unittest_endpoint, deployment_data):
    deployment_id = deployment_data['id']
    update_setting_url = '{}/deployments/{}/settings/'.format(unittest_endpoint, deployment_id)
    deployment_url = '{}/deployments/{}/'.format(unittest_endpoint, deployment_id)
    status_url = '{}/status_url'.format(unittest_endpoint)
    responses.add(
        responses.PATCH,
        update_setting_url,
        headers={'Location': status_url},
        status=202)
    responses.add(
        responses.GET,
        status_url,
        headers={'Location': deployment_url},
        status=303)


@responses.activate
@pytest.mark.usefixtures('deployment_get_response')
def test_cannot_update_nothing(deployment_data):
    deployment_id = deployment_data['id']
    deployment = Deployment.get(deployment_id)
    with pytest.raises(ValueError):
        deployment.update_drift_tracking_settings()


@responses.activate
@pytest.mark.usefixtures('deployment_get_response', 'deployment_settings_get_response')
def test_get_drift_tracking_settings(unittest_endpoint, deployment_data,
                                     deployment_settings_response_data):
    deployment_id = deployment_data['id']
    deployment = Deployment.get(deployment_id)
    settings = deployment.get_drift_tracking_settings()

    response = from_api(deployment_settings_response_data)
    expected_settings = {'target_drift': response['target_drift'],
                         'feature_drift': response['feature_drift']}
    assert settings == expected_settings


@responses.activate
@pytest.mark.usefixtures('deployment_get_response', 'deployment_settings_update_response')
@pytest.mark.parametrize('param, expected_request', [
    ({'target_drift_enabled': True}, {'targetDrift': {'enabled': True}}),
    ({'feature_drift_enabled': False}, {'featureDrift': {'enabled': False}}),
    ({'target_drift_enabled': True, 'feature_drift_enabled': True},
     {'targetDrift': {'enabled': True}, 'featureDrift': {'enabled': True}}),
    ({'target_drift_enabled': False, 'feature_drift_enabled': False},
     {'targetDrift': {'enabled': False}, 'featureDrift': {'enabled': False}})
])
def test_update_drift_tracking_settings(unittest_endpoint, deployment_data,
                                        param, expected_request):
    deployment_id = deployment_data['id']
    deployment = Deployment.get(deployment_id)
    deployment.update_drift_tracking_settings(**param)

    request_body = request_body_to_json(responses.calls[1].request)
    assert request_body == expected_request


@responses.activate
@pytest.mark.usefixtures('deployment_get_response', 'deployment_settings_get_response')
def test_get_association_id_settings(unittest_endpoint, deployment_data,
                                     deployment_settings_response_data):
    deployment_id = deployment_data['id']
    deployment = Deployment.get(deployment_id)
    settings = deployment.get_association_id_settings()

    response = from_api(deployment_settings_response_data)
    expected_settings = {
        'column_names': response['association_id']['column_names'],
        'required_in_prediction_requests': response[
            'association_id']['required_in_prediction_requests']
    }
    assert settings == expected_settings


@responses.activate
@pytest.mark.usefixtures('deployment_get_response', 'deployment_settings_update_response')
@pytest.mark.parametrize('param, expected_request', [
    ({'column_names': ['a']}, {'associationId': {'columnNames': ['a']}}),
    ({'required_in_prediction_requests': True},
     {'associationId': {'requiredInPredictionRequests': True}}),
    ({'column_names': ['a'], 'required_in_prediction_requests': True},
     {'associationId': {'columnNames': ['a'], 'requiredInPredictionRequests': True}}),
])
def test_update_association_id_settings(unittest_endpoint, deployment_data,
                                        param, expected_request):
    deployment_id = deployment_data['id']
    deployment = Deployment.get(deployment_id)
    deployment.update_association_id_settings(**param)

    request_body = request_body_to_json(responses.calls[1].request)
    assert request_body == expected_request


@responses.activate
@pytest.mark.usefixtures('deployment_get_response')
def test_get_prediction_warning_settings(unittest_endpoint, deployment_data,
                                         deployment_settings_response_data):
    deployment_id = deployment_data['id']
    url = '{}/deployments/{}/settings/'.format(unittest_endpoint, deployment_id)
    responses.add(
        responses.GET,
        url,
        body=json.dumps(deployment_settings_response_data),
        status=200,
        content_type='application/json')

    deployment = Deployment.get(deployment_id)
    settings = deployment.get_prediction_warning_settings()

    response = from_api(deployment_settings_response_data)
    assert settings == response['prediction_warning']


@responses.activate
@pytest.mark.usefixtures('deployment_get_response')
@pytest.mark.parametrize(
    'param, expected_request',
    [
        (
            {
                'prediction_warning_enabled': True,
                'use_default_boundaries': False,
                'upper_boundary': 1337,
                'lower_boundary': 0,
            },
            {
                'predictionWarning': {
                    'enabled': True,
                    'customBoundaries': {'upper': 1337, 'lower': 0},
                }
            },
        ),
        (
            {'prediction_warning_enabled': True, 'use_default_boundaries': True},
            {'predictionWarning': {'enabled': True, 'customBoundaries': None}},
        ),
        ({'prediction_warning_enabled': False}, {'predictionWarning': {'enabled': False}}),
    ],
)
def test_update_prediction_warning_settings(unittest_endpoint, deployment_data,
                                            param, expected_request):
    deployment_id = deployment_data['id']
    update_setting_url = '{}/deployments/{}/settings/'.format(unittest_endpoint, deployment_id)
    deployment_url = '{}/deployments/{}/'.format(unittest_endpoint, deployment_id)
    status_url = '{}/status_url'.format(unittest_endpoint)
    responses.add(
        responses.PATCH,
        update_setting_url,
        headers={'Location': status_url},
        status=202)
    responses.add(
        responses.GET,
        status_url,
        headers={'Location': deployment_url},
        status=303)
    deployment = Deployment.get(deployment_id)
    deployment.update_prediction_warning_settings(**param)
    request_body = request_body_to_json(responses.calls[1].request)
    assert request_body == expected_request


@responses.activate
@pytest.mark.usefixtures('deployment_get_response')
def test_get_predictions_data_collection_settings(
    unittest_endpoint, deployment_data, deployment_settings_response_data
):
    deployment_id = deployment_data['id']
    url = '{}/deployments/{}/settings/'.format(unittest_endpoint, deployment_id)
    responses.add(
        responses.GET,
        url,
        body=json.dumps(deployment_settings_response_data),
        status=200,
        content_type='application/json',
    )

    deployment = Deployment.get(deployment_id)
    settings = deployment.get_predictions_data_collection_settings()

    response = from_api(deployment_settings_response_data)
    assert settings == response['predictions_data_collection']


@responses.activate
@pytest.mark.usefixtures('deployment_get_response')
@pytest.mark.parametrize(
    'param, expected_request',
    [
        ({'enabled': True}, {'predictionsDataCollection': {'enabled': True}}),
        ({'enabled': False}, {'predictionsDataCollection': {'enabled': False}}),
    ],
)
def test_update_predictions_data_collection_settings(
    unittest_endpoint, deployment_data, param, expected_request
):
    deployment_id = deployment_data['id']
    update_setting_url = '{}/deployments/{}/settings/'.format(
        unittest_endpoint, deployment_id
    )
    deployment_url = '{}/deployments/{}/'.format(unittest_endpoint, deployment_id)
    status_url = '{}/status_url'.format(unittest_endpoint)
    responses.add(
        responses.PATCH,
        update_setting_url,
        headers={'Location': status_url},
        status=202,
    )
    responses.add(
        responses.GET, status_url, headers={'Location': deployment_url}, status=303
    )
    deployment = Deployment.get(deployment_id)
    deployment.update_predictions_data_collection_settings(**param)

    request_body = request_body_to_json(responses.calls[1].request)
    assert request_body == expected_request


@responses.activate
@pytest.mark.usefixtures('deployment_get_response')
def test_get_prediction_intervals_settings(unittest_endpoint, deployment_data,
                                           deployment_settings_response_data):
    deployment_id = deployment_data['id']
    url = '{}/deployments/{}/settings/'.format(unittest_endpoint, deployment_id)
    responses.add(
        responses.GET,
        url,
        body=json.dumps(deployment_settings_response_data),
        status=200,
        content_type='application/json')

    deployment = Deployment.get(deployment_id)
    settings = deployment.get_prediction_intervals_settings()

    response = from_api(deployment_settings_response_data)
    expected_settings = response['prediction_intervals']
    assert settings == expected_settings


@responses.activate
@pytest.mark.usefixtures('deployment_get_response')
def test_update_prediction_intervals_settings(unittest_endpoint, deployment_data):
    deployment_id = deployment_data['id']
    update_setting_url = '{}/deployments/{}/settings/'.format(unittest_endpoint, deployment_id)
    deployment_url = '{}/deployments/{}/'.format(unittest_endpoint, deployment_id)
    status_url = '{}/status_url'.format(unittest_endpoint)
    project_id = deployment_data['model']['projectId']
    model_id = deployment_data['model']['id']

    # responses for ensuring percentile is calculated for model
    responses.add(
        responses.POST,
        '{}/projects/{}/models/{}/predictionIntervals/'.format(
            unittest_endpoint, project_id, model_id
        ),
        status=202,
        headers={'Location': '{}/projects/{}/jobs/1/'.format(unittest_endpoint, project_id)}
    )
    responses.add(
        responses.GET,
        '{}/projects/{}/jobs/1/'.format(unittest_endpoint, project_id),
        status=303,
        content_type='application/json',
        body=json.dumps({
            'status': u'COMPLETED', 'model_id': model_id, 'is_blocked': False,
            'url': '{}/projects/{}/jobs/1/'.format(unittest_endpoint, project_id),
            'job_type': 'calculate_prediction_intervals', 'project_id': project_id, 'id': 1
        }),
        headers={'Location': '{}/projects/{}/jobs/1/result/'.format(unittest_endpoint, project_id)}
    )

    # responses for updating deployment settings
    responses.add(
        responses.PATCH,
        update_setting_url,
        headers={'Location': status_url},
        status=202)
    responses.add(
        responses.GET,
        status_url,
        headers={'Location': deployment_url},
        status=303)

    deployment = Deployment.get(deployment_id)
    deployment.update_prediction_intervals_settings(percentiles=[70], enabled=False)

    request_body = request_body_to_json(responses.calls[5].request)
    assert request_body == {'predictionIntervals': {'enabled': False, 'percentiles': [70]}}



from collections import OrderedDict

import dateutil
import trafaret as t

from datarobot.models.api_object import APIObject
from ..enums import SERVICE_STAT_METRIC
from ..helpers.deployment_monitoring import DeploymentQueryBuilderMixin
from ..utils import encode_utf8_if_py2, from_api


class ServiceStats(APIObject, DeploymentQueryBuilderMixin):
    """Deployment service stats information.

    Attributes
    ----------
    model_id : str
        the model used to retrieve service stats metrics
    period : dict
        the time period used to retrieve service stats metrics
    metrics : dict
        the service stats metrics
    """

    _path = 'deployments/{}/serviceStats/'
    _period = t.Dict({
        t.Key('start'): t.String >> (lambda s: dateutil.parser.parse(s)),
        t.Key('end'): t.String >> (lambda s: dateutil.parser.parse(s)),
    })
    _converter = t.Dict({
        t.Key('period'): _period,
        t.Key('metrics'): t.Mapping(t.Enum(*SERVICE_STAT_METRIC.ALL), t.Int | t.Float),
        t.Key('model_id'): t.String() | t.Null,
    }).allow_extra('*')

    def __init__(self, period=None, metrics=None, model_id=None):
        self.period = period or {}
        self.metrics = metrics or {}
        self.model_id = model_id

    def __repr__(self):
        return encode_utf8_if_py2(u'{}({} | {} - {})'.format(
            self.__class__.__name__, self.model_id,
            self.period.get('start'), self.period.get('end')))

    def __getitem__(self, item):
        return self.metrics.get(item)

    @classmethod
    def get(cls, deployment_id, model_id=None, start_time=None, end_time=None,
            execution_time_quantile=None, response_time_quantile=None,
            slow_requests_threshold=None):
        """Retrieve value of service stat metrics over a certain time period.

        .. versionadded:: v2.18

        Parameters
        ----------
        deployment_id : str
            the id of the deployment
        model_id : str, optional
            the id of the model
        start_time : datetime, optional
            start of the time period
        end_time : datetime, optional
            end of the time period
        execution_time_quantile : float, optional
            quantile for `executionTime`, defaults to 0.5
        response_time_quantile : float, optional
            quantile for `responseTime`, defaults to 0.5
        slow_requests_threshold : float, optional
            threshold for `slowRequests`, defaults to 1000

        Returns
        -------
        service_stats : ServiceStats
            the queried service stats metrics
        """

        path = cls._path.format(deployment_id)
        params = {
            'start_time': start_time,
            'end_time': end_time,
            'model_id': model_id,
            'execution_time_quantile': execution_time_quantile,
            'response_time_quantile': response_time_quantile,
            'slow_requests_threshold': slow_requests_threshold,
        }
        params = cls._build_query_params(**params)
        data = cls._client.get(path, params=params).json()

        # we don't want to convert keys of the metrics object
        metrics = data.pop('metrics')

        data = from_api(data, keep_null_keys=True)
        data['metrics'] = metrics
        return cls.from_data(data)


class ServiceStatsOverTime(APIObject, DeploymentQueryBuilderMixin):
    """Deployment service stats over time information.

        Attributes
        ----------
        model_id : str
            the model used to retrieve accuracy metric
        metric : str
            the service stat metric being retrieved
        buckets : dict
            how the service stat metric changes over time
        summary : dict
            summary for the service stat metric
        """

    _path = 'deployments/{}/serviceStatsOverTime/'
    _period = t.Dict({
        t.Key('start'): t.String >> (lambda s: dateutil.parser.parse(s)),
        t.Key('end'): t.String >> (lambda s: dateutil.parser.parse(s)),
    })
    _bucket = t.Dict({
        t.Key('period'): _period | t.Null,
        t.Key('value'): t.Int | t.Float | t.Null,
    }).allow_extra('*')
    _converter = t.Dict({
        t.Key('buckets'): t.List(_bucket),
        t.Key('summary'): _bucket,
        t.Key('metric'): t.String(),
        t.Key('model_id'): t.String() | t.Null,
    }).allow_extra('*')

    def __init__(self, buckets=None, summary=None, metric=None, model_id=None):
        self.buckets = buckets
        self.summary = summary
        self.metric = metric
        self.model_id = model_id

    def __repr__(self):
        period = self.summary.get('period') or {}
        return encode_utf8_if_py2(u'{}({} | {} | {} - {})'.format(
            self.__class__.__name__, self.model_id, self.metric,
            period.get('start'), period.get('end')))

    @classmethod
    def get(cls, deployment_id, metric=None, model_id=None, start_time=None, end_time=None,
            bucket_size=None, quantile=None, threshold=None):
        """Retrieve information about how a service stat metric changes over a certain time period.

        .. versionadded:: v2.18

        Parameters
        ----------
        deployment_id : str
            the id of the deployment
        metric : SERVICE_STAT_METRIC, optional
            the service stat metric to retrieve
        model_id : str, optional
            the id of the model
        start_time : datetime, optional
            start of the time period
        end_time : datetime, optional
            end of the time period
        bucket_size : str, optional
            time duration of a bucket, in ISO 8601 time duration format
        quantile : float, optional
            quantile for 'executionTime' or 'responseTime', ignored when querying other metrics
        threshold : int, optional
            threshold for 'slowQueries', ignored when querying other metrics

        Returns
        -------
        service_stats_over_time : ServiceStatsOverTime
            the queried service stat over time information
        """

        path = cls._path.format(deployment_id)
        params = {
            'start_time': start_time,
            'end_time': end_time,
            'model_id': model_id,
            'metric': metric,
            'bucket_size': bucket_size,
            'quantile': quantile,
            'threshold': threshold,
        }
        params = cls._build_query_params(**params)
        data = cls._client.get(path, params=params).json()
        case_converted = from_api(data, keep_null_keys=True)
        return cls.from_data(case_converted)

    @property
    def bucket_values(self):
        """The metric value for all time buckets, keyed by start time of the bucket.

        Returns
        -------
        bucket_values: OrderedDict
        """

        values = [(bucket['period']['start'], bucket['value']) for bucket in self.buckets]
        return OrderedDict(values)



import datetime

import pytest
import responses
from tests.utils import request_body_to_json

from datarobot import FeatureEngineeringGraph, SharingAccess


@pytest.fixture
def create_feature_engineering_graph_items():
    return [
        {
            'identifier': 'user',
            'catalogVersionId': '5c88a37770fc42a2fcc62759',
            'snapshotPolicy': 'latest'
        },
        {
            'identifier': 'transaction',
            'catalogVersionId': '5c88a37770fc42a2fcc62760',
            'primaryTemporalKey': 'date',
            'snapshotPolicy': 'latest'
        },
    ]


@pytest.fixture
def create_feature_engineering_graph_relationships():
    return [
        {
            'table2Identifier': 'user',
            'table1Keys': ['user_id', 'dept_id'],
            'table2Keys': ['user_id', 'dept_id'],
        },
        {
            'table1Identifier': 'user',
            'table2Identifier': 'transaction',
            'table1Keys': ['user_id'],
            'table2Keys': ['user_id'],
        },
    ]


@pytest.fixture
def feature_engineering_graph_response(create_feature_engineering_graph_items,
                                       create_feature_engineering_graph_relationships):
    """ Payload to create a Feature Source via API
    """
    return {
        'id': '5a530498d5c1f302d6d176c8',
        'name': 'Customer Features',
        'description': 'Credit card customer features (using activity in last 2 months)',
        'tableDefinitions': create_feature_engineering_graph_items,
        'relationships': create_feature_engineering_graph_relationships,
        'featureDerivationWindowStart': -14,
        'featureDerivationWindowEnd': -1,
        'timeUnit': 'DAY',
        'created': '2016-07-26T02:29:58.546312Z',
        'lastModified': '2016-07-26T02:29:58.546312Z',
        'creatorFullName': 'abc@datarobot.com',
        'modifierFullName': 'abc@datarobot.com',
        'creatorUserId': '5a530498d5c1f302d6d176c5',
        'lastModifiedUserId': '5a530498d5c1f302d6d176c5',
        'numberOfProjects': 2,
        'linkageKeys': ['user_id', 'dept_id'],
    }


@pytest.fixture
def list_feature_engineering_graph_response(create_feature_engineering_graph_items,
                                            create_feature_engineering_graph_relationships):
    return {
        'data': [
            {
                'id': '5a530498d5c1f302d6d176c8',
                'name': 'Customer Features',
                'description': 'Credit card customer features (using activity in last 2 months)',
                'tableDefinitions': create_feature_engineering_graph_items,
                'relationships': create_feature_engineering_graph_relationships,
                'featureDerivationWindowStart': -14,
                'featureDerivationWindowEnd': -1,
                'timeUnit': 'DAY',
                'created': '2016-07-26T02:29:58.546312Z',
                'lastModified': '2016-07-26T02:29:58.546312Z',
                'creatorFullName': 'abc@datarobot.com',
                'modifierFullName': 'abc@datarobot.com',
                'creatorUserId': '5a530498d5c1f302d6d176c5',
                'lastModifiedUserId': '5a530498d5c1f302d6d176c5',
                'numberOfProjects': 2,
                'linkageKeys': ['user_id', 'dept_id'],
            },
            {
                'id': '5a530498d5c1f302d6d175c8',
                'name': 'Profile Features',
                'description': 'Profile feature',
                'tableDefinitions': create_feature_engineering_graph_items,
                'relationships': create_feature_engineering_graph_relationships,
                'featureDerivationWindowStart': -10,
                'featureDerivationWindowEnd': -1,
                'timeUnit': 'HOUR',
                'created': '2016-07-26T02:29:58.546312Z',
                'lastModified': '2016-07-26T02:29:58.546312Z',
                'creatorFullName': 'abc@datarobot.com',
                'modifierFullName': 'abc@datarobot.com',
                'creatorUserId': '5a530498d5c1f302d6d176c5',
                'lastModifiedUserId': '5a530498d5c1f302d6d176c5',
                'numberOfProjects': 2,
                'linkageKeys': ['user_id', 'p_id'],
            },
        ]
    }


@responses.activate
def test_feature_engineering_graph_creation(feature_engineering_graph_response,
                                            create_feature_engineering_graph_items,
                                            create_feature_engineering_graph_relationships):
    responses.add(responses.POST,
                  'https://host_name.com/featureEngineeringGraphs/',
                  json=feature_engineering_graph_response)

    feature_engineering_graph = FeatureEngineeringGraph()
    description = 'Credit card customer features (using activity in last 2 months)'
    created_feg = feature_engineering_graph.create(
        name='Customer Features',
        description=description,
        table_definitions=create_feature_engineering_graph_items,
        relationships=create_feature_engineering_graph_relationships,
        time_unit='DAY',
        feature_derivation_window_start=-14,
        feature_derivation_window_end=-1)
    assert responses.calls[0].request.method == 'POST'
    assert created_feg.feature_engineering_graph_id == '5a530498d5c1f302d6d176c8'
    assert created_feg.name == 'Customer Features'
    assert created_feg.description == description
    assert len(created_feg.table_definitions) == 2
    assert len(created_feg.relationships) == 2
    assert created_feg.time_unit == feature_engineering_graph_response['timeUnit']
    assert created_feg.feature_derivation_window_start == -14
    assert created_feg.feature_derivation_window_end == -1
    assert isinstance(created_feg.created, datetime.datetime)
    assert isinstance(created_feg.last_modified, datetime.datetime)
    assert isinstance(created_feg.created, datetime.datetime)


@responses.activate
def test_feature_engineering_graph_replace(feature_engineering_graph_response,
                                           create_feature_engineering_graph_items,
                                           create_feature_engineering_graph_relationships):
    responses.add(responses.PUT,
                  'https://host_name.com/featureEngineeringGraphs/F-ID/',
                  status=200,
                  json=feature_engineering_graph_response)

    feature_engineering_graph = FeatureEngineeringGraph()
    description = 'Credit card customer features (using activity in last 2 months)'
    updated_feg = feature_engineering_graph.replace(
        id='F-ID',
        name='Customer Features',
        description=description,
        table_definitions=create_feature_engineering_graph_items,
        relationships=create_feature_engineering_graph_relationships,
        time_unit='DAY',
        feature_derivation_window_start=-14,
        feature_derivation_window_end=-1,
        is_draft=False)
    assert responses.calls[0].request.method == 'PUT'
    payload = request_body_to_json(responses.calls[0].request)
    assert payload['name'] == 'Customer Features'
    assert len(payload['tableDefinitions']) == 2
    assert len(payload['relationships']) == 2
    assert payload['timeUnit'] == 'DAY'
    assert payload['featureDerivationWindowStart'] == -14
    assert payload['featureDerivationWindowEnd'] == -1

    assert updated_feg.feature_engineering_graph_id == '5a530498d5c1f302d6d176c8'
    assert updated_feg.name == 'Customer Features'
    assert updated_feg.description == description
    assert len(updated_feg.table_definitions) == 2
    assert len(updated_feg.relationships) == 2
    assert updated_feg.time_unit == 'DAY'
    assert updated_feg.feature_derivation_window_start == -14
    assert updated_feg.feature_derivation_window_end == -1
    assert isinstance(updated_feg.created, datetime.datetime)
    assert isinstance(updated_feg.last_modified, datetime.datetime)


@responses.activate
def test_feature_engineering_graph_updation():
    responses.add(responses.PATCH,
                  'https://host_name.com/featureEngineeringGraphs/F-ID/',
                  status=200,
                  json={'name': 'new-name', 'description': 'new-descrip'})

    feature_engineering_graph = FeatureEngineeringGraph('F-ID')
    feature_engineering_graph.update(
        name='new-name',
        description='new-descrip')

    assert responses.calls[0].request.method == 'PATCH'
    payload = request_body_to_json(responses.calls[0].request)
    assert payload['name'] == 'new-name'
    assert payload['description'] == 'new-descrip'


@responses.activate
def test_feature_engineering_graph_retrieve(feature_engineering_graph_response):
    graph_id = feature_engineering_graph_response['id']
    responses.add(
        responses.GET,
        'https://host_name.com/featureEngineeringGraphs/{}/'.format(graph_id),
        json=feature_engineering_graph_response
    )

    feature_engineering_graph = FeatureEngineeringGraph.get(feature_engineering_graph_id=graph_id)

    assert responses.calls[0].request.method == 'GET'
    assert feature_engineering_graph.feature_engineering_graph_id == graph_id
    assert feature_engineering_graph.name == 'Customer Features'
    assert len(feature_engineering_graph.table_definitions) == 2
    assert len(feature_engineering_graph.relationships) == 2
    assert feature_engineering_graph.time_unit == 'DAY'
    assert feature_engineering_graph.feature_derivation_window_start == -14
    assert feature_engineering_graph.feature_derivation_window_end == -1
    assert isinstance(feature_engineering_graph.created, datetime.datetime)
    assert isinstance(feature_engineering_graph.last_modified, datetime.datetime)
    assert isinstance(feature_engineering_graph.created, datetime.datetime)


@responses.activate
def test_feature_engineering_graph_list(list_feature_engineering_graph_response):
    responses.add(responses.GET,
                  'https://host_name.com/featureEngineeringGraphs/',
                  json=list_feature_engineering_graph_response)

    graphs = FeatureEngineeringGraph.list()

    for graph, server_payload in zip(graphs, list_feature_engineering_graph_response['data']):
        assert graph.feature_engineering_graph_id == server_payload['id']
        assert graph.name == server_payload['name']
        assert graph.description == server_payload['description']
        assert len(graph.table_definitions) == 2
        assert len(graph.relationships) == 2
        fdw = server_payload['featureDerivationWindowStart']
        assert graph.feature_derivation_window_start == fdw
        assert graph.feature_derivation_window_end == server_payload['featureDerivationWindowEnd']
        assert graph.time_unit == server_payload['timeUnit']
        assert graph.creator_full_name == server_payload['creatorFullName']
        assert graph.modifier_full_name == server_payload['modifierFullName']
        assert graph.number_of_projects == server_payload['numberOfProjects']
        assert graph.linkage_keys == server_payload['linkageKeys']


@responses.activate
def test_delete(feature_engineering_graph_response):
    graph_id = feature_engineering_graph_response['id']
    responses.add(
        responses.GET,
        'https://host_name.com/featureEngineeringGraphs/{}/'.format(graph_id),
        json=feature_engineering_graph_response
    )
    delete_url = 'https://host_name.com/featureEngineeringGraphs/{}/'.format(graph_id)
    responses.add(responses.DELETE, delete_url)

    feature_engineering_graph = FeatureEngineeringGraph.get(graph_id)
    feature_engineering_graph.delete()

    assert responses.calls[1].request.method == responses.DELETE
    assert responses.calls[1].request.url == delete_url


@responses.activate
def test_share(feature_engineering_graph_response):
    graph = FeatureEngineeringGraph.from_server_data(feature_engineering_graph_response)
    request = SharingAccess('me@datarobot.com', 'EDITOR')
    responses.add(responses.PATCH,
                  'https://host_name.com/featureEngineeringGraphs/{}/accessControls/'.
                  format(graph.feature_engineering_graph_id),
                  status=200)

    graph.share([request])

    actual_payload = request_body_to_json(responses.calls[0].request)
    assert actual_payload == {'permissions': [{'username': 'me@datarobot.com', 'role': 'EDITOR'}]}


@responses.activate
def test_remove(feature_engineering_graph_response):
    graph = FeatureEngineeringGraph.from_server_data(feature_engineering_graph_response)
    request = SharingAccess('me@datarobot.com', None)
    responses.add(responses.PATCH,
                  'https://host_name.com/featureEngineeringGraphs/{}/accessControls/'.
                  format(graph.feature_engineering_graph_id),
                  status=200)

    graph.share([request])

    actual_payload = request_body_to_json(responses.calls[0].request)
    assert actual_payload == {'permissions': [{'username': 'me@datarobot.com', 'role': None}]}


@responses.activate
def test_get_access_list(feature_engineering_graph_response):
    graph = FeatureEngineeringGraph.from_server_data(feature_engineering_graph_response)
    access_list = {
        'data': [
            {'username': 'me@datarobot.com',
             'userId': '1234deadbeeffeeddead4321',
             'role': 'OWNER',
             'canShare': True
             }
        ],
        'count': 1,
        'previous': None,
        'next': None
    }

    access_record = access_list['data'][0]
    responses.add(responses.GET,
                  'https://host_name.com/featureEngineeringGraphs/{}/accessControls/'.
                  format(graph.feature_engineering_graph_id),
                  json=access_list)

    response = graph.get_access_list()
    assert len(response) == 1

    share_info = response[0]
    assert share_info.username == access_record['username']
    assert share_info.user_id == access_record['userId']
    assert share_info.role == access_record['role']
    assert share_info.can_share == access_record['canShare']
